import logging
import os
import shutil

import h5py
import numpy as np

from wormpose import default_config
from wormpose.dataset.features import calculate_crop_window_size
from wormpose.dataset.loader import load_dataset
from wormpose.pose.postures_model import PosturesModel
from wormpose.images.synthetic_real_comparison import SyntheticRealComparison
from wormpose.pose.worm_skeleton_utils import skeletons_to_angles

logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def _calibrate_one_video(frame_preprocessing, frames_dataset, video_name,
                         features_dataset, results_dir, image_shape,
                         max_frames, theta_dims):
    """
    Evaluate image metric score on labelled frames.
    """
    features = features_dataset[video_name]
    labelled_angles = skeletons_to_angles(features.skeletons, theta_dims=theta_dims)

    skel_is_not_nan = ~np.any(np.isnan(features.skeletons), axis=(1, 2))
    labelled_indexes = np.where(skel_is_not_nan)[0]

    synthetic_real_comparison = SyntheticRealComparison(frame_preprocessing=frame_preprocessing,
                                                        nb_skeleton_joints=len(features.skeletons[0]),
                                                        template_measurements=features.measurements,
                                                        image_shape=image_shape)
    all_scores = []
    all_synth = []
    all_real = []

    with frames_dataset.open(video_name) as frames:
        frames_amount = min(max_frames, len(labelled_indexes))

        random_label_index = np.random.choice(labelled_indexes, frames_amount, replace=False)
        angles = labelled_angles[random_label_index]

        for theta, index in zip(angles, random_label_index):
            cur_frame = frames[index]

            score, _ = synthetic_real_comparison.run(theta=theta,
                                                     synth_template_skeleton=features.skeletons[index],
                                                     synth_template_frame=cur_frame,
                                                     real_frame_orig=cur_frame)
            all_real.append(np.array(synthetic_real_comparison.last_real_image))
            all_synth.append(np.array(synthetic_real_comparison.last_synth_image))
            all_scores.append(score)

    results_file = os.path.join(results_dir, video_name + '_calibration.h5')
    with h5py.File(results_file, "w") as f:
        f.create_dataset('real_images', data=all_real)
        f.create_dataset('synth_images', data=all_synth)
        f.create_dataset('scores', data=all_scores)

    logger.info("Evaluated known skeletons reconstruction for {},"
                " average score {:.4f}".format(video_name, np.mean(all_scores)))
    return results_file


def calibrate(dataset_loader,
              dataset_path,
              calibration_num_frames=None,
              work_dir=None,
              theta_dims=None):
    if calibration_num_frames is None:
        calibration_num_frames = 500
    if work_dir is None:
        work_dir = default_config.WORK_DIR
    if theta_dims is None:
        theta_dims = PosturesModel().dimensions

    logger.info("Calibrating with parameters: \n\t" + ', \n\t'.join(['{}: {}'.format(x[0], x[1])
                                                                     for x in {'dataset_loader': dataset_loader,
                                                                               'dataset_path': os.path.abspath(dataset_path),
                                                                               'calibration_num_frames': calibration_num_frames,
                                                                               'work_dir': os.path.abspath(work_dir),
                                                                               'theta_dims': theta_dims,
                                                                               }.items()]))

    dataset_root_name = os.path.basename(os.path.normpath(dataset_path))
    project_dir = os.path.join(work_dir, dataset_root_name)
    calibration_results_dir = os.path.join(project_dir, default_config.CALIBRATION_RESULTS_DIR)
    os.makedirs(calibration_results_dir, exist_ok=True)

    features_dataset, frames_dataset, frame_preprocessing = load_dataset(dataset_loader, dataset_path)
    image_shape = calculate_crop_window_size(features_dataset)
    video_names = frames_dataset.video_names()

    for video_name in video_names:
        results_file = _calibrate_one_video(frame_preprocessing=frame_preprocessing,
                                            frames_dataset=frames_dataset,
                                            video_name=video_name,
                                            features_dataset=features_dataset,
                                            results_dir=calibration_results_dir,
                                            image_shape=image_shape,
                                            max_frames=calibration_num_frames,
                                            theta_dims=theta_dims)
        yield video_name, results_file


def main():
    import argparse

    parser = argparse.ArgumentParser()

    # inputs
    parser.add_argument('dataset_loader', type=str)
    parser.add_argument('dataset_path', type=str)
    parser.add_argument('--calibration_num_frames', type=int,
                        help="How many frames to perform the calibration in order to evaluate the image metric")
    parser.add_argument('--theta_dims', type=int)
    parser.add_argument('--work_dir', type=str,
                        help="Root folder for all experiments")
    args = parser.parse_args()

    list(calibrate(**vars(args)))


if __name__ == "__main__":
    main()