import numpy as np


class Features(object):
    def __init__(self, features_dict):

        #  features
        skeletons = features_dict['skeletons']
        head_width = features_dict['head_width']
        midbody_width = features_dict['midbody_width']
        tail_width = features_dict['tail_width']
        frame_rate = features_dict['frame_rate']
        # optional features
        ventral_side = features_dict['ventral_side'] if 'ventral_side' in features_dict else None
        timestamp = features_dict['timestamp'] if 'timestamp' in features_dict else None

        # validate features
        if not (len(skeletons) == len(head_width) == len(midbody_width) == len(tail_width)):
            raise ValueError("inconsistent features")
        if not(head_width.dtype == midbody_width.dtype == tail_width.dtype) \
                or not np.issubdtype(head_width.dtype, np.floating):
            raise TypeError("Body measurements type should be identical and floating point")
        if not np.issubdtype(skeletons.dtype, np.floating):
            raise TypeError("Skeleton type should be floating point")
        if len(skeletons.shape) != 3 or skeletons.shape[2] != 2:
            raise ValueError("Wrong skeleton shape")
        if skeletons.shape[1] < 20:
            raise UserWarning("Low number of skeleton joints (< 20), consider interpolating to improve quality of synthetic images")

        self.skeletons = skeletons
        self.frame_rate = frame_rate
        self.ventral_side = ventral_side if ventral_side is not None else 'anticlockwise'
        self.timestamp = timestamp

        if self.timestamp is None:  # default behavior : all frames are equidistant in time
            self.timestamp = np.arange(0, len(self.skeletons))

        # calculate length from skeleton
        worm_length = np.full(len(skeletons), np.nan, dtype=head_width.dtype)
        where_skel_nan = np.where(~np.any(np.isnan(skeletons), axis=(1, 2)))
        for i in where_skel_nan[0]:
            skel = skeletons[i]
            worm_length[i] = np.sum(np.sqrt(np.sum((skel[:-1] - skel[1:]) ** 2, axis=1)))

        self.measurements = np.stack([worm_length, head_width, midbody_width, tail_width], axis=1)
        self.measurements.dtype = {'names': ('worm_length', 'head_width', 'midbody_width', 'tail_width'),
                                   'formats': (worm_length.dtype, head_width.dtype, midbody_width.dtype, tail_width.dtype)}


def calculate_crop_window_size(features):
    """
    Returns an image shape that is just big enough to view the worm object,
    the image will be cropped (or expanded) to that size as an input to the neural network
    Can be overriden in child classes for another behavior (for example a fixed chosen size)
    :return: A tuple of two integer values (height, width)
    """
    import math
    MINIMUM_CROP_SIZE = 32

    # calculate the image shape as maximum of all the average worm lengths for each video
    # then we know the longest worm will fit the image shape,
    # the smaller ones will just have more background space
    max_average_worm_length = np.nanmax([np.nanmean(x.measurements['worm_length']) for x in features.values()])
    crop_size = max(MINIMUM_CROP_SIZE, int(max_average_worm_length))

    # round to even number
    crop_size = math.ceil(float(crop_size) / 2) * 2
    return crop_size, crop_size
