import pkg_resources

from wormpose.dataset.base_dataset import BaseFramesDataset, BaseFeaturesDataset, BaseFramePreprocessing
from wormpose.dataset.features import Features

_loaders_register = {}


def _register_loaders():

    for entry_point in pkg_resources.iter_entry_points('worm_dataset_loaders'):
        module_name = entry_point.name
        module = entry_point.load()
        _loaders_register[module_name] = module.FramesDataset, module.FeaturesDataset, module.FramePreprocessing


def load_dataset(loader_name, dataset_path):

    _register_loaders()

    if loader_name not in _loaders_register:
        raise NotImplementedError("Dataset loader: \'{}\' not found, "
                                  "did you run setup.py?".format(loader_name))

    return _load_dataset(dataset_path, *_loaders_register[loader_name])


def _load_dataset(dataset_path, frames_dataset_class, features_dataset_class, frame_preprocessing_class):

    frames_dataset = frames_dataset_class(dataset_path)
    video_names = frames_dataset.video_names()

    if len(video_names) == 0:
        raise ValueError("Video names list is empty, no video to analyze in {}".format(dataset_path))
    if not all(isinstance(n, str) for n in video_names):
        raise NotImplementedError("Only video names as list of strings supported type.")
    if len(video_names) > len(set(video_names)):
        raise ValueError("Video names from FramesDataset must be unique but duplicates were found: {}.".format(video_names))

    raw_features_dataset = features_dataset_class(dataset_path, video_names)
    features_dataset = {}
    for video_name in video_names:
        features_dataset[video_name] = Features(raw_features_dataset.get_features(video_name))

    frame_preprocessing = frame_preprocessing_class()

    return features_dataset, frames_dataset, frame_preprocessing

