"""
Loaders for Amolf tracker data

The dataset_path is the path of one h5 file
"""
import os

import h5py
import numpy as np

from wormpose.dataset.base_dataset import BaseFramesDataset, BaseFeaturesDataset
from wormpose.dataset.image_processing.simple_frame_preprocessing import SimpleFramePreprocessing
from wormpose.pose.worm_skeleton_utils import get_joint_indexes

SEPARATOR_CHAR = ':'


class FramesDataset(BaseFramesDataset):
    class FramesReader(object):

        def __init__(self, dataset_path, video_name):
            self.h5_filename = dataset_path
            self.video_name = video_name

        def __enter__(self):
            self.f = h5py.File(self.h5_filename, 'r')
            frames = self.f
            for k in self.video_name:
                frames = frames[k]

            return frames['grayWormImage']

        def __exit__(self, exc_type, exc_value, traceback):
            self.f.close()

    def __init__(self, dataset_path):

        if not os.path.isfile(dataset_path):
            raise FileNotFoundError()

        self.dataset_path = dataset_path
        with h5py.File(self.dataset_path, 'r') as f:

            all_names = []

            def print_attrs(name, _):
                all_names.append(name)

            f.visititems(print_attrs)
            all_names_split = [n.split('/') for n in all_names]
            all_valid_names = list(filter(lambda x: len(x) == 4 and any(i.isdigit() for i in x[-1]), all_names_split))

            if any([SEPARATOR_CHAR in x for x in all_valid_names]):
                raise ValueError("need to choose another separator")

        self.videos_paths = {SEPARATOR_CHAR.join(n): n for n in all_valid_names}

    def video_names(self):
        return list(sorted(self.videos_paths.keys()))

    def open(self, video_name):
        return self.FramesReader(self.dataset_path, self.videos_paths[video_name])


def _read_features(dataset_path, name):

    with h5py.File(dataset_path, 'r') as f:
        framerate = f['video']['frameRate'][0]

        worm_dset = f
        for k in name.split(SEPARATOR_CHAR):
            worm_dset = worm_dset[k]

        skeletons = worm_dset['skeletonSpline'][()]
        width = worm_dset['skeletonSplineWidth'][()]

        # filter all zeros and bad frames
        bad_frames_filter = [np.all(skeletons == 0, axis=(1, 2))]
        if 'badFrames' in worm_dset:
            bad_frames_filter += [worm_dset['badFrames'][()]]
        if 'orientationFixed' in worm_dset:
            bad_frames_filter += [~worm_dset['orientationFixed'][()]]
        if 'Collision' in worm_dset:
            bad_frames_filter += [worm_dset['Collision'][()]]
        invalid_frames = np.logical_or.reduce(bad_frames_filter)
        skeletons[invalid_frames] = np.nan
        width[invalid_frames] = np.nan

        head_joint, mid_body_joint, tail_body_joint = get_joint_indexes(len(width[0]))

        head_width = width[:, head_joint]
        midbody_width = width[:, mid_body_joint]
        tail_width = width[:, tail_body_joint]

        return {'skeletons': skeletons,
                'head_width': head_width,
                'midbody_width': midbody_width,
                'tail_width': tail_width,
                'frame_rate': framerate}


class FeaturesDataset(BaseFeaturesDataset):

    def __init__(self, dataset_path, video_names):

        self.dataset_path = dataset_path

        self._features = {}

        for video_name in video_names:
            self._features[video_name] = _read_features(dataset_path, video_name)

    def get_features(self, video_name):
        return self._features[video_name]


class FramePreprocessing(SimpleFramePreprocessing):

    def run(self, frame):

        # first crop black area on bottom right
        where_zero = np.where(frame > 0)
        if len(where_zero[0]) > 0 and len(where_zero[1]) > 0:
            frame = frame[:np.max(where_zero[0]), :np.max(where_zero[1])]

        return super().run(frame)


