"""
Loaders for Tierpsy tracker data

The files should be organized this way :

+-- dataset_path
|   +-- video_name0
|   |   +-- video_name0.hdf5
|   |   +-- video_name0_features.hdf5  or video_name0_featuresN.hdf5
|   +-- video_name1
|   |   +-- video_name1.hdf5
|   |   +-- video_name1_features.hdf5 or video_name1_featuresN.hdf5
|   ...

WARNING not all configurations of Tierpsy tracker are supported.
"""

import glob
import json
import logging
import os
import time
from xml.dom import minidom

import h5py
import numpy as np

from wormpose.dataset.base_dataset import BaseFramesDataset
from wormpose.dataset.base_dataset import BaseFeaturesDataset
from wormpose.dataset.image_processing.simple_frame_preprocessing import SimpleFramePreprocessing

logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class FramesDataset(BaseFramesDataset):

    class FramesReader(object):

        def __init__(self, h5_filename):
            self.h5_filename = h5_filename

        def __enter__(self):
            self.f = h5py.File(self.h5_filename, 'r')
            return self.f['mask']

        def __exit__(self, exc_type, exc_value, traceback):
            self.f.close()

    def __init__(self, dataset_path):

        if not os.path.isdir(dataset_path):
            raise FileNotFoundError("'{}' needs to be a directory containing Tierpsy videos.".format(dataset_path))

        self.videos_paths = {}

        for video_name in os.listdir(dataset_path):
            if os.path.isdir(os.path.join(dataset_path, video_name)):
                video_path = os.path.join(dataset_path, video_name, video_name + '.hdf5')
                if os.path.exists(video_path):
                    self.videos_paths[video_name] = video_path

        if len(self.videos_paths) == 0:
            raise FileNotFoundError("Couldn't find Tierpsy videos in '{}'.".format(dataset_path))

    def video_names(self):
        return list(self.videos_paths.keys())

    def open(self, video_name):
        return self.FramesReader(self.videos_paths[video_name])


def _read_features(root_dir, name):
    start = time.time()

    h5file = os.path.join(root_dir, name, name + '.hdf5')
    h5featurefile = glob.glob(os.path.join(root_dir, name, name + '*features*.hdf5'))[0]

    with h5py.File(h5file, 'r') as f:

        len_frames = len(f['mask'])

        # Getting the frame rate
        mask_attrs = f['mask'].attrs
        if 'fps' in mask_attrs:
            frame_rate = mask_attrs['fps']
        elif 'expected_fps' in mask_attrs:
            frame_rate = mask_attrs['expected_fps']
        else:
            raise ValueError("Frame rate not found in file.")

        # Trying to get the ventral side attribute (optional)
        try:
            ventral_side = json.loads(f['experiment_info'][()].decode('utf-8'))['ventral_side']
        except:
            ventral_side = None

        frames_timestamp = f['timestamp']['raw'][:]
        try:
            stage_position_pix = f['stage_position_pix'][:]
        except:
            stage_position_pix = np.zeros((len(frames_timestamp), 2), dtype=int)

        try:
            xml_info = f['xml_info'][()]
            xml = minidom.parseString(xml_info)
            microns = xml.getElementsByTagName('microns')[0].getElementsByTagName('x')[0].firstChild.nodeValue
            pixels = xml.getElementsByTagName('pixels')[0].getElementsByTagName('x')[0].firstChild.nodeValue
            ratio_microns_pixel = abs(float(microns)) / abs(float(pixels))
        except:
            ratio_microns_pixel = 1.

    with h5py.File(h5featurefile, 'r') as f:
        skeletons = f['coordinates']['skeletons'][:]

        try:
            features_timeseries = f[list(filter(lambda x: 'timeseries' in x, f))[0]][:]
            feature_names = features_timeseries.dtype.names
        except:
            raise ValueError("Feature time series not found in file.")

        # only use one worm (smallest index)
        worm_indexes = features_timeseries['worm_index']
        worm_index = np.min(worm_indexes)
        cur_worm = np.where(worm_indexes == worm_index)[0]
        features_timeseries = features_timeseries[cur_worm]

        features_timestamp = features_timeseries['timestamp'].astype(int)

        if 'trajectories_data' in f:
            trajectories_data = f['trajectories_data']
            skeletons_id = trajectories_data['skeleton_id']
            skeletons_timestamp = np.zeros(len(skeletons), dtype=np.int)
            for i in range(len(skeletons)):
                skeletons_timestamp[i] = np.where(skeletons_id == i)[0][0]
        else:
            skeletons_timestamp = features_timestamp

        head_width_name = list(filter(lambda x: 'head' in x and 'width' in x, feature_names))[0]
        head_width = features_timeseries[head_width_name] * ratio_microns_pixel
        midbody_width_name = list(filter(lambda x: 'midbody' in x and 'width' in x, feature_names))[0]
        midbody_width = features_timeseries[midbody_width_name] * ratio_microns_pixel
        tail_width_name = list(filter(lambda x: 'tail' in x and 'width' in x, feature_names))[0]
        tail_width = features_timeseries[tail_width_name] * ratio_microns_pixel
        measurements = np.stack([head_width, midbody_width, tail_width], axis=1)

    resampled_skeletons = np.full((len_frames,) + skeletons.shape[1:], np.nan, dtype=np.float32)
    resampled_measurements = np.full((len_frames,) + measurements.shape[1:], np.nan, dtype=np.float32)

    def match_indexes(x, y):
        sorted_index = np.searchsorted(x, y)
        yindex = np.take(np.arange(len(x)), sorted_index, mode="clip")
        mask = x[yindex] != y
        result = np.ma.array(yindex, mask=mask)
        return result

    feature_indexes = match_indexes(features_timestamp, frames_timestamp)
    skeleton_indexes = match_indexes(skeletons_timestamp, frames_timestamp)

    for frame_index in range(len_frames):

        if not skeleton_indexes.mask[frame_index]:
            cur_skel = skeletons[skeleton_indexes[frame_index]]

            if not np.any(np.isnan(cur_skel)) and not np.any(np.isnan(stage_position_pix[frame_index])):
                cur_skel = (cur_skel * ratio_microns_pixel - stage_position_pix[frame_index]).astype(int)
                resampled_skeletons[frame_index] = cur_skel

        if not feature_indexes.mask[frame_index]:
            cur_measurements = measurements[feature_indexes[frame_index]]
            resampled_measurements[frame_index] = cur_measurements

    end = time.time()
    logger.info("Preprocessed file {} in {:.1f}s".format(name, end - start))

    skeletons = resampled_skeletons
    head_width = resampled_measurements[:, 0]
    midbody_width = resampled_measurements[:, 1]
    tail_width = resampled_measurements[:, 2]
    timestamp = frames_timestamp

    return {'skeletons': skeletons,
            'head_width': head_width,
            'midbody_width': midbody_width,
            'tail_width': tail_width,
            'frame_rate': frame_rate,
            'ventral_side': ventral_side,
            'timestamp': timestamp}


class FeaturesDataset(BaseFeaturesDataset):

    def __init__(self, dataset_path, video_names):

        self._features = {}
        for video_name in video_names:
            self._features[video_name] = _read_features(dataset_path, video_name)

    def get_features(self, video_name):
        return self._features[video_name]


FramePreprocessing = SimpleFramePreprocessing