import numpy as np

from wormpose.dataset.features import calculate_crop_window_size
from wormpose.dataset.loader import load_dataset
from wormpose.images.synthetic_dataset import SyntheticDataset
from wormpose.pose.postures_model import PosturesModel


class SyntheticSimpleVisualizer(object):
    """
    Utility class to visualize the synthetic images
    """

    def __init__(self, dataset_loader, dataset_path, postures_model=None, video_name=None):

        features_dataset, frames_dataset, frame_preprocessing = load_dataset(dataset_loader, dataset_path)

        if postures_model is None:
            postures_model = PosturesModel()
        if video_name is None:
            video_name = frames_dataset.video_names()[0]

        features = features_dataset[video_name]
        self.skeletons = features.skeletons
        self.measurements = features.measurements

        self.output_image_shape = calculate_crop_window_size(features_dataset)

        self.synthetic_dataset = SyntheticDataset(frame_preprocessing=frame_preprocessing,
                                                  output_image_shape=self.output_image_shape,
                                                  enable_random_augmentations=False)
        skel_is_not_nan = ~np.any(np.isnan(self.skeletons), axis=(1, 2))
        self.labelled_indexes = np.where(skel_is_not_nan)[0]
        if len(self.labelled_indexes) == 0:
            raise ValueError("No template frames found in the dataset, can't generate synthetic images.")
        self.frames_dataset = frames_dataset
        self.video_name = video_name
        self.postures_model = postures_model

    def generate(self):
        out_image = np.empty(self.output_image_shape, dtype=np.uint8)

        postures_model_gen = self.postures_model.generate()
        with self.frames_dataset.open(self.video_name) as frames:
            while True:
                theta = next(postures_model_gen)
                random_label_index = np.random.choice(self.labelled_indexes)
                self.synthetic_dataset.generate(theta=theta,
                                                template_skeleton=self.skeletons[random_label_index],
                                                template_frame=frames[random_label_index],
                                                out_image=out_image,
                                                worm_measurements=self.measurements)
                yield out_image, theta


def main():
    import argparse
    import cv2

    parser = argparse.ArgumentParser()
    parser.add_argument('dataset_loader', type=str)
    parser.add_argument('dataset_path', type=str)
    parser.add_argument('--video_name', type=str)
    args = parser.parse_args()

    synth_visualizer_gen = SyntheticSimpleVisualizer(args.dataset_loader, args.dataset_path).generate()

    while True:
        synth_image, _ = next(synth_visualizer_gen)

        cv2.imshow("synth_image", synth_image)
        cv2.waitKey()


if __name__ == "__main__":
    main()
