import json
import logging
import os
import shutil
import tempfile
import time
import uuid

from wormpose import default_config
from wormpose.dataset.features import calculate_crop_window_size
from wormpose.dataset.loader import load_dataset
from wormpose.machine_learning import eval_data_generator, synthetic_data_generator
from wormpose.pose.postures_model import PosturesModel

logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def generate(dataset_loader,
             dataset_path,
             num_train_samples=None,
             num_eval_samples=None,
             num_process=None,
             scratch_dir=None,
             work_dir=None,
             disable_image_augmentation=None,
             postures_model=None):
    import multiprocessing as mp
    mp.set_start_method('spawn', force=True)

    if num_process is None:
        num_process = os.cpu_count()
    if scratch_dir is None:
        scratch_dir = tempfile.gettempdir()
    if num_train_samples is None:
        num_train_samples = int(5e5)
    if num_eval_samples is None:
        num_eval_samples = int(1e4)
    if work_dir is None:
        work_dir = default_config.WORK_DIR
    if postures_model is None:
        postures_model = PosturesModel()
    if disable_image_augmentation is None:
        disable_image_augmentation= False

    logger.info("Generating training data with parameters: \n\t" + ', \n\t'.join(['{}: {}'.format(x[0], x[1])
                                                                                  for x in
                                                                                  {'dataset_path': os.path.abspath(dataset_path),
                                                                                   'dataset_loader': dataset_loader,
                                                                                   'num_process': num_process,
                                                                                   'scratch_dir': os.path.abspath(scratch_dir),
                                                                                   'num_train_samples': num_train_samples,
                                                                                   'num_eval_samples': num_eval_samples,
                                                                                   'work_dir': os.path.abspath(work_dir),
                                                                                   'postures_model': postures_model,
                                                                                   'disable_image_augmentation': disable_image_augmentation,
                                                                                   }.items()]))

    dataset_root_name = os.path.basename(os.path.normpath(dataset_path))

    # setup folders
    if not os.path.exists(work_dir):
        os.mkdir(work_dir)
    project_dir = os.path.join(work_dir, dataset_root_name)
    os.makedirs(project_dir, exist_ok=True)

    tfrecords_dataset_root = os.path.join(project_dir, default_config.TRAINING_DATA_DIR)
    if os.path.exists(tfrecords_dataset_root):
        shutil.rmtree(tfrecords_dataset_root)
    temp_dir = os.path.join(scratch_dir, str(uuid.uuid4()))
    if os.path.exists(temp_dir):
        shutil.rmtree(temp_dir)
    os.mkdir(temp_dir)
    logger.info('Storing temporary files in : {}'.format(temp_dir))

    features_dataset, frames_dataset, frame_preprocessing = load_dataset(dataset_loader, dataset_path)
    image_shape = calculate_crop_window_size(features_dataset)

    start = time.time()
    gen = synthetic_data_generator.generate(frame_preprocessing=frame_preprocessing,
                                            frames_dataset=frames_dataset,
                                            features_dataset=features_dataset,
                                            num_process=num_process,
                                            dataset_out_dir=temp_dir,
                                            num_samples=num_train_samples,
                                            temp_dir=temp_dir,
                                            image_shape=image_shape,
                                            postures_model=postures_model,
                                            disable_image_augmentation=disable_image_augmentation,
                                            file_extension="tfrecord",
                                            filename_prefix="train")
    for progress in gen:
        yield progress
    yield 1.

    num_eval_samples = eval_data_generator.generate(frame_preprocessing=frame_preprocessing,
                                 frames_dataset=frames_dataset,
                                 features_dataset=features_dataset,
                                 image_shape=image_shape,
                                 num_samples=num_eval_samples,
                                 dataset_out_dir=temp_dir,
                                 theta_dims=postures_model.dimensions)

    shutil.copytree(temp_dir, tfrecords_dataset_root)
    # saving infos about data generation
    config = {'dataset_loader': dataset_loader,
              'image_shape': image_shape,
              'theta_dimensions': postures_model.dimensions,
              'num_train_samples': num_train_samples,
              'num_eval_samples': num_eval_samples,
              }
    config_path = os.path.join(project_dir, default_config.CONFIG_FILENAME)
    with open(config_path, 'w') as f:
        json.dump(config, f, indent=4)

    end = time.time()
    logger.info('Done generating training data in : {:.1f}s, all files at: {}'.format(end - start, os.path.abspath(tfrecords_dataset_root)))
    logger.info("Saved configuration file at: {}".format(os.path.abspath(config_path)))


def main():
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('dataset_loader', type=str)
    parser.add_argument('dataset_path', type=str)
    parser.add_argument('--num_train_samples', type=int,
                        help="How many training samples to generate")
    parser.add_argument('--num_eval_samples', type=int,
                        help="How many evaluation samples to generate")
    parser.add_argument('--scratch_dir', type=str,
                        help="Where to store temporary intermediate results")
    parser.add_argument('--work_dir', type=str,
                        help="Root folder for all experiments")
    parser.add_argument('--num_process', type=int,
                        help="How many worker processes")
    parser.add_argument('--disable_image_augmentation', action='store_true',
                        help="If set, disable image augmentation in training images")
    args = parser.parse_args()

    last_progress = None
    for progress in generate(**vars(args)):
        prog_percent = int(progress * 100)
        if prog_percent != last_progress:
            logger.info("Generating training data: {}% done".format(prog_percent))
        last_progress = prog_percent


if __name__ == "__main__":
    main()
