import os
import pickle
from multiprocessing import Process, Manager

import numpy as np

from wormpose.images.synthetic_real_comparison import SyntheticRealComparison


def run(frame_preprocessing, temp_dir, num_process, theta_two_options, video_name, frames_dataset, features_dataset,
        image_shape):
    import logging
    logging.basicConfig()
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.DEBUG)

    manager = Manager()

    data_reading_queue = manager.Queue()
    results_queue = manager.Queue()

    features = features_dataset[video_name]

    skel_is_not_nan = ~np.any(np.isnan(features.skeletons), axis=(1, 2))
    template_indexes = np.where(skel_is_not_nan)[0]
    if len(template_indexes) == 0:
        logger.error("Can't calculate image metric, there is no labelled frame in the video to use as a template, "
                     "stopping analysis.")
        return np.full((len(features.skeletons), 2), np.nan), \
               np.full(((len(features.skeletons), 2) + features.skeletons.shape[1:]), np.nan)

    workers = [Process(target=_compare_pred_real, args=(frame_preprocessing,
                                                        data_reading_queue,
                                                        results_queue,
                                                        features.skeletons,
                                                        features.measurements,
                                                        image_shape,
                                                        temp_dir))
               for _ in range(num_process)]
    for w in workers:
        w.start()

    _produce_data_chunks(num_process, data_reading_queue, temp_dir, video_name, frames_dataset, template_indexes, theta_two_options)

    for w in workers:
        w.join()

    results_files = []
    while not results_queue.empty():
        results_files.append(results_queue.get())
    results_files = sorted(results_files)

    matching_scores = []
    matching_skeletons = []
    for result_filename in results_files:
        with open(result_filename, 'rb') as f:
            results = pickle.load(f)
        matching_scores.append(results['scores'])
        matching_skeletons.append(results['skeletons'])
        os.remove(result_filename)

    return np.concatenate(matching_scores), np.concatenate(matching_skeletons)


def _produce_data_chunks(num_process, data_reading_queue, temp_dir, video_name, frames_dataset, template_indexes, theta_two_options, chunk_size=100):

    def save_chunk(chunk_data, chunk_index, real_frames, synth_template_frames):
        compare_data_filename = os.path.join(temp_dir, "compare_data_{}.pkl".format(chunk_index))
        with open(compare_data_filename, 'wb') as f:
            pickle.dump({'chunk_data': chunk_data, 'real_frames': real_frames, 'template_frames': synth_template_frames}, f)
        data_reading_queue.put((compare_data_filename, chunk_index))

    with frames_dataset.open(video_name) as frames:

        chunk_data = []
        chunk_index = 0
        chunk_start_frame_index = 0
        chunk_synth_templates = []

        for frame_index, (theta, theta_flip) in enumerate(theta_two_options):

            synth_template_index = template_indexes[np.abs(template_indexes - frame_index).argmin()]
            chunk_synth_templates.append(frames[synth_template_index])

            chunk_data.append((theta, theta_flip, synth_template_index))

            if len(chunk_data) >= chunk_size:
                save_chunk(chunk_data, chunk_index, frames[chunk_start_frame_index:frame_index + 1],
                           chunk_synth_templates)
                chunk_start_frame_index = frame_index + 1
                chunk_index += 1
                chunk_data = []
                chunk_synth_templates = []

        if len(chunk_data) > 0:
            save_chunk(chunk_data, chunk_index, frames[chunk_start_frame_index:frame_index + 1], chunk_synth_templates)

    for _ in range(num_process):
        data_reading_queue.put(None)


def _compare_pred_real(frame_preprocessing, data_reading_queue, results_queue, skeletons,
                       measurements, image_shape, temp_dir):

    synthetic_real_comparison = SyntheticRealComparison(frame_preprocessing=frame_preprocessing,
                                                        nb_skeleton_joints=len(skeletons[0]),
                                                        template_measurements=measurements,
                                                        image_shape=image_shape)
    while True:
        queue_data = data_reading_queue.get()
        if queue_data is None:
            break

        compare_data_filename, chunk_index = queue_data

        with open(compare_data_filename, 'rb') as f:
            compare_data = pickle.load(f)
            chunk_data = compare_data['chunk_data']
            chunk_real_frames = compare_data['real_frames']
            chunk_template_frames = compare_data['template_frames']
        os.remove(compare_data_filename)

        all_scores = []
        all_skel = []
        for data, real_frame_orig, synth_template_frame in zip(chunk_data, chunk_real_frames, chunk_template_frames):
            theta, theta_flip, synth_template_index = data

            res = [synthetic_real_comparison.run(theta=x,
                                                 synth_template_skeleton=skeletons[synth_template_index],
                                                 synth_template_frame=synth_template_frame,
                                                 real_frame_orig=real_frame_orig)
                   for x in [theta, theta_flip]]

            all_scores.append([res[0][0], res[1][0]])
            all_skel.append([res[0][1], res[1][1]])

        out_filename = os.path.join(temp_dir, "compare_results_{:09d}.pkl".format(chunk_index))

        with open(out_filename, 'wb') as f:
            pickle.dump({'scores': all_scores, 'skeletons': all_skel}, f)

        results_queue.put(out_filename)
