import math
import random

import cv2
import numpy as np

from wormpose.dataset.image_processing import frame_preprocessor
from wormpose.images.worm_drawing import WormOutlineDrawing
from wormpose.pose.worm_skeleton_utils import get_joint_indexes


def _make_synth_worm_measurements(nb_skeleton_joints, measurements):
    head_width = np.nanmean(measurements['head_width'])
    midbody_width = np.nanmean(measurements['midbody_width'])
    tail_width = np.nanmean(measurements['tail_width'])

    head_joint, mid_body_joint, tail_body_joint = get_joint_indexes(nb_skeleton_joints)
    worm_thickness = np.empty(nb_skeleton_joints, np.float32)
    worm_thickness[:head_joint] = head_width
    worm_thickness[head_joint:mid_body_joint] = np.linspace(head_width, midbody_width,
                                                            mid_body_joint - head_joint)
    worm_thickness[mid_body_joint:tail_body_joint] = np.linspace(midbody_width, tail_width,
                                                                 tail_body_joint - mid_body_joint)
    worm_thickness[tail_body_joint:] = tail_width

    worm_length = np.nanmean(measurements['worm_length'])

    return worm_length, worm_thickness


def _make_calc_one_segment_coords():
    skel_perp = np.empty((2,), dtype=np.float32)

    def run(joint_a, joint_b, width, dest):
        perp = [-(joint_b[1] - joint_a[1]), joint_b[0] - joint_a[0]]
        norm = math.sqrt(perp[0] * perp[0] + perp[1] * perp[1])

        skel_perp[0] = (perp[0] / norm) * (width / 2)
        skel_perp[1] = (perp[1] / norm) * (width / 2)

        dest[0] = joint_a + skel_perp
        dest[1] = joint_b + skel_perp
        dest[2] = joint_b - skel_perp
        dest[3] = joint_a - skel_perp

    return run


def _make_calc_all_coords(nb_skeleton_joints, enable_random_augmentations):
    step = max(1, nb_skeleton_joints // SyntheticDataset.BODY_SEGMENTS_DIVIDER)
    last = nb_skeleton_joints - step

    old_coords = np.empty((last + 2, 4, 2), dtype=np.float32)
    new_coords = np.empty((last + 2, 4, 2), dtype=np.float32)
    calc_one_segment_coords = _make_calc_one_segment_coords()

    def run(cur_skel, new_skel, cur_worm_thickness):
        # before head
        old_width = cur_worm_thickness[0]
        calc_one_segment_coords(joint_a=cur_skel[0] + cur_skel[0] - cur_skel[step],
                                joint_b=cur_skel[0],
                                width=old_width,
                                dest=old_coords[0])
        calc_one_segment_coords(joint_a=new_skel[0] + new_skel[0] - new_skel[step],
                                joint_b=new_skel[0],
                                width=old_width,
                                dest=new_coords[0])
        # after tail
        old_width = cur_worm_thickness[last]
        calc_one_segment_coords(joint_a=cur_skel[last - 1 + step],
                                joint_b=cur_skel[last - 1 + step] + cur_skel[last - 1 + step] - cur_skel[last - 1],
                                width=old_width,
                                dest=old_coords[1])
        calc_one_segment_coords(joint_a=new_skel[last - 1 + step],
                                joint_b=new_skel[last - 1 + step] + new_skel[last - 1 + step] - new_skel[last - 1],
                                width=old_width,
                                dest=new_coords[1])

        if enable_random_augmentations and np.random.random() >= SyntheticDataset.DRAW_TAIL_ON_TOP_PROBABILITY:
            # draw the worm starting from the head (the tail will be on top)
            joint_indexes = range(0, last)
        else:
            # draw starting from the tail (the head will be on top)
            joint_indexes = reversed(range(0, last))

        for index, joint_index in enumerate(joint_indexes):
            old_width = cur_worm_thickness[joint_index]
            calc_one_segment_coords(joint_a=cur_skel[joint_index],
                                    joint_b=cur_skel[joint_index + step],
                                    width=old_width,
                                    dest=old_coords[index + 2])
            calc_one_segment_coords(joint_a=new_skel[joint_index],
                                    joint_b=new_skel[joint_index + step],
                                    width=old_width,
                                    dest=new_coords[index + 2])

        return old_coords, new_coords

    return run


class SyntheticDataset(object):
    # synthetic shape generator params
    BODY_SEGMENTS_DIVIDER = 16  # draw worm segments of a length 1/16 of the worm length

    # data augmentation parameters (applied when enable_random_augmentations is True)
    BLUR_PROBABILITY = 0.25  # extra blurring happens 25% of the time
    DRAW_TAIL_ON_TOP_PROBABILITY = 0.5  # draw tail on top of head 50% of the time
    MAX_OFFSET_PERCENT = 0.05  # extra translation max 5% of image shape

    def __init__(self, frame_preprocessing,
                 enable_random_augmentations,
                 output_image_shape):

        self.output_image_shape = (output_image_shape[0], output_image_shape[1])
        self.enable_random_augmentations = enable_random_augmentations

        self.frame_preprocessing = frame_preprocessing

        self.max_offset = (int(self.output_image_shape[0] * self.MAX_OFFSET_PERCENT),
                           int(self.output_image_shape[1] * self.MAX_OFFSET_PERCENT))

        self.nb_skeleton_joints = None
        self.calc_all_coords = None
        self.synth_worm_measurements_cache = {}

        self.worm_outline_drawing = WormOutlineDrawing()

        self.frames_infos_cache = {}

        # blurring image when random augmentations apply :
        # blur kernel is chosen between 3 and 13 (or 10% of the image size)
        self.gaussian_blur_filter_sizes = np.arange(3, min(14, int(0.10 * min(output_image_shape))), 2) \
            if enable_random_augmentations else None

        # pre allocate numpy arrays for calculations
        self.out_image = np.empty(self.output_image_shape, dtype=np.float32)
        self.source_image_mask = None
        self.target_image_mask = np.empty(self.output_image_shape, dtype=np.uint8)
        self.new_shape_outline_image = np.empty(self.output_image_shape, dtype=np.uint8)
        self.target_image = np.empty(self.output_image_shape, dtype=np.uint8)
        self.cur_alpha = np.empty(self.output_image_shape, dtype=np.float32)
        self.all_alpha = np.empty(self.output_image_shape, dtype=np.float32)

    def _get_recentered_frame(self, template_frame, template_skeleton):

        frame_id = hash(template_frame.tostring())

        if frame_id in self.frames_infos_cache:
            worm_roi, bg_mean_color = self.frames_infos_cache[frame_id]
        else:
            frame, bg_mean_color, worm_roi = frame_preprocessor.run(self.frame_preprocessing, template_frame)
            self.frames_infos_cache[frame_id] = (worm_roi, bg_mean_color)

        template_skeletons_recentered = template_skeleton - (worm_roi[1].start, worm_roi[0].start)
        return template_frame[worm_roi], bg_mean_color, template_skeletons_recentered

    def _get_synth_measurements(self, worm_measurements):

        measurements_id = id(worm_measurements)

        if measurements_id not in self.synth_worm_measurements_cache:
            self.synth_worm_measurements_cache[measurements_id] = _make_synth_worm_measurements(self.nb_skeleton_joints,
                                                                                                worm_measurements)

        return self.synth_worm_measurements_cache[measurements_id]

    def generate(self, theta, template_frame, template_skeleton, worm_measurements, out_image):

        if out_image.dtype != np.uint8:
            raise NotImplementedError("Only uint8 images supported.")

        # get the amount of translation that will be applied
        tx, ty = (np.random.uniform(-self.max_offset[0], self.max_offset[0]),
                  np.random.uniform(-self.max_offset[1], self.max_offset[1])) \
            if self.enable_random_augmentations else (0, 0)

        # get image and skeleton recentered to region of interest
        recentered_frame, cur_bg_color, template_skel = self._get_recentered_frame(template_frame, template_skeleton)

        if self.nb_skeleton_joints is None:
            self.nb_skeleton_joints = len(template_skel)

        if len(template_skel) != self.nb_skeleton_joints:
            raise NotImplementedError("Can't change the amount of skeleton joints at runtime.")

        # get and augment if needed worm length and worm thickness
        cur_worm_length, cur_worm_thickness = self._get_synth_measurements(worm_measurements)
        thickness_multiplier = 1.2 if not self.enable_random_augmentations \
            else np.random.uniform(1.1, 1.3)
        cur_worm_thickness = cur_worm_thickness * thickness_multiplier
        if self.enable_random_augmentations:
            cur_worm_length = cur_worm_length * np.random.uniform(0.9, 1.1)

        # calculate the new skeleton coordinates, and also draw the new worm shape outline
        self.new_shape_outline_image.fill(0)
        new_skel = self.worm_outline_drawing.visualize_skeleton(theta=theta,
                                                                worm_length=cur_worm_length,
                                                                worm_thickness=cur_worm_thickness / 2,
                                                                output_image=self.new_shape_outline_image,
                                                                translation_offset=(tx, ty))

        self.out_image.fill(0.)

        # preallocate mask, grow it if necessary
        if self.source_image_mask is None \
                or self.source_image_mask.shape[0] < template_frame.shape[0] \
                or self.source_image_mask.shape[1] < template_frame.shape[1]:
            self.source_image_mask = np.empty_like(template_frame)
        source_image_mask = self.source_image_mask[:recentered_frame.shape[0], :recentered_frame.shape[1]]

        # calculate all transformations of body segments first
        if self.calc_all_coords is None:
            self.calc_all_coords = _make_calc_all_coords(nb_skeleton_joints=len(template_skel),
                                                         enable_random_augmentations=self.enable_random_augmentations)
        all_old_coords, all_new_coords = self.calc_all_coords(template_skel, new_skel, cur_worm_thickness)

        for old_coords, new_coords in zip(all_old_coords, all_new_coords):
            transform = cv2.getAffineTransform(old_coords[:3], new_coords[:3])

            source_image_mask.fill(0)
            cv2.fillConvexPoly(source_image_mask, old_coords.astype(np.int), 255)
            masked_frame = cv2.bitwise_and(recentered_frame, source_image_mask)

            cv2.warpAffine(masked_frame, transform, dsize=self.output_image_shape,
                           borderMode=cv2.BORDER_CONSTANT, flags=cv2.INTER_NEAREST, dst=self.target_image)

            self.target_image_mask.fill(0)
            cv2.fillConvexPoly(self.target_image_mask, new_coords.astype(np.int), 255)

            cv2.bitwise_and(self.target_image, self.target_image_mask, dst=self.target_image)

            np.greater(self.target_image, 0, out=self.cur_alpha)
            np.greater(self.out_image, 0, out=self.all_alpha)

            self.all_alpha += self.cur_alpha
            self.all_alpha[self.all_alpha > 1] = 0.5

            self.out_image += self.target_image
            self.out_image *= self.all_alpha

        self.out_image[self.new_shape_outline_image == 0] = 0
        self.out_image[self.out_image == 0] = cur_bg_color
        cv2.medianBlur(self.out_image, 3, dst=self.out_image)

        if self.enable_random_augmentations and np.random.uniform() > 1 - self.BLUR_PROBABILITY:
            ksize = random.choice(self.gaussian_blur_filter_sizes)
            cv2.GaussianBlur(self.out_image, (ksize, ksize), 0, dst=self.out_image)

        np.copyto(out_image, self.out_image, casting='unsafe')

        return cur_bg_color, new_skel
