import numpy as np

from wormpose.images import image_scoring
from wormpose.images.image_scoring import fit_bounding_box_to_synth_image
from wormpose.images.real_dataset import RealDataset
from wormpose.images.synthetic_dataset import SyntheticDataset


class SyntheticRealComparison(object):
    CROP_PADDING = 2

    def __init__(self, frame_preprocessing, nb_skeleton_joints, template_measurements, image_shape):

        self.real_dataset = RealDataset(frame_preprocessing=frame_preprocessing,
                                        output_image_shape=image_shape)

        self.worm_measurements = template_measurements
        self.synthetic_dataset = SyntheticDataset(frame_preprocessing=frame_preprocessing,
                                                  output_image_shape=image_shape,
                                                  enable_random_augmentations=False)
        self.last_synth_image = np.empty(image_shape, np.uint8)
        self.last_real_image = None

    def run(self, theta, synth_template_skeleton, synth_template_frame, real_frame_orig):
        self.last_real_image, skel_offset = self.real_dataset.process_frame(real_frame_orig)
        cur_bg_color, synth_skel = self.synthetic_dataset.generate(theta,
                                                                   template_frame=synth_template_frame,
                                                                   template_skeleton=synth_template_skeleton,
                                                                   out_image=self.last_synth_image,
                                                                   worm_measurements=self.worm_measurements)

        # Crop the synthetic image to the object of interest before doing the image comparison,
        # we don't need the full image with all the background, still keep a little padding around the worm.
        left, right, bottom, top = fit_bounding_box_to_synth_image(self.last_synth_image, cur_bg_color)
        np.subtract(synth_skel, (bottom, left), out=synth_skel)
        cropped_synth_image = self.last_synth_image[left:right, bottom:top]

        # Perform the image comparison between the real image and the reconstructed synthetic image cropped.
        # This gives a heat map, we are interested in the maximum value of the heatmap and its location.
        # This maximum score gives an estimation of the confidence of the prediction.
        score, score_loc = image_scoring.template_match(source_image=self.last_real_image, template_image=cropped_synth_image)

        # Using the heatmap maximum coordinates, we can transform the coordinates of the reconstructed skeleton
        # to the original image coordinates system.
        synth_skel += score_loc
        synth_skel += np.array([skel_offset[0], skel_offset[1]])

        return score, synth_skel
