import cv2
import numpy as np

from wormpose.pose.worm_skeleton_utils import get_joint_indexes


def calculate_backbone_points(theta, worm_length):
    """
     Calculates (x,y) coordinates of the worm backbone from the tangent angle (theta) vector
     """
    n_backbone_points = len(theta)
    backbone_section_length = worm_length / (n_backbone_points - 1)

    rx = np.cumsum(backbone_section_length * np.cos(theta))
    ry = np.cumsum(backbone_section_length * np.sin(theta))

    return rx, ry


def calculate_backbone_points_centered(theta, worm_length, output_image_shape):

    rx, ry = calculate_backbone_points(theta, worm_length)

    # center the skeleton coordinates in the middle of the output image
    rx = rx - np.min(rx) + (output_image_shape[1] - (np.max(rx) - np.min(rx))) / 2
    ry = ry - np.min(ry) + (output_image_shape[0] - (np.max(ry) - np.min(ry))) / 2

    return rx, ry


def _make_draw_worm_body():

    polygon = [[]] * 4
    orthogonal = np.empty((2,), dtype=np.float32)
    points_todraw = np.empty((4, 2), dtype=int)

    def run(worm_thickness, img, rx, ry):

        n_backbone_points = len(worm_thickness)
        body_color = 255

        # draw two circles at the extremities
        cv2.circle(img, (int(rx[0]), int(ry[0])), int(worm_thickness[0]), body_color, -1, 8)
        cv2.circle(img, (int(rx[n_backbone_points - 1]), int(ry[n_backbone_points - 1])),
                   int(worm_thickness[n_backbone_points - 1]), body_color, -1, 8)

        # draw convex quadrilateral polygons in the middle
        prev_boundary_0 = None
        prev_boundary_1 = None

        for i in range(2, n_backbone_points, 2):
            cur_backbone = np.array([rx[i], ry[i]])
            prev_backbone = np.array([rx[i - 2], ry[i - 2]])

            orthogonal[0] = cur_backbone[1] - prev_backbone[1]
            orthogonal[1] = -(cur_backbone[0] - prev_backbone[0])

            length = np.sqrt(orthogonal[0] * orthogonal[0] + orthogonal[1] * orthogonal[1])
            if not np.isclose(length, 0):
                orthogonal[0] /= length
                orthogonal[1] /= length

            polygon[0] = cur_backbone + orthogonal * worm_thickness[i]
            polygon[1] = cur_backbone - orthogonal * worm_thickness[i]
            if i == 2:
                polygon[2] = prev_backbone - orthogonal * worm_thickness[i - 1]
                polygon[3] = prev_backbone + orthogonal * worm_thickness[i - 1]
            else:
                polygon[3] = prev_boundary_0
                polygon[2] = prev_boundary_1

            prev_boundary_0 = polygon[0]
            prev_boundary_1 = polygon[1]

            points_todraw[0] = polygon[0]
            points_todraw[1] = polygon[1]
            points_todraw[2] = polygon[2]
            points_todraw[3] = polygon[3]
            cv2.fillConvexPoly(img, points_todraw, body_color, 8)

    return run


class WormOutlineDrawing(object):

    def __init__(self):
        self.draw_worm_body = _make_draw_worm_body()

    def visualize_skeleton(self, theta, worm_length, worm_thickness, output_image, translation_offset=(0, 0)):

        rx, ry = calculate_backbone_points_centered(theta, worm_length, output_image.shape)
        rx += translation_offset[0]
        ry += translation_offset[1]

        if len(rx) != len(worm_thickness):
            rx = np.interp(np.linspace(0, len(rx) - 1, len(worm_thickness)), np.arange(0, len(rx)), rx)
            ry = np.interp(np.linspace(0, len(ry) - 1, len(worm_thickness)), np.arange(0, len(ry)), ry)

        self.draw_worm_body(worm_thickness, output_image, rx, ry)
        return np.stack([rx, ry], axis=1)


def draw_skeleton(image, skeleton, color, head_color):

    if np.any(np.isnan(skeleton)):
        return

    skeleton = skeleton.astype(int)
    for skel_index in range(len(skeleton) - 1):
        cv2.line(image,
                 (skeleton[skel_index][0], skeleton[skel_index][1]),
                 (skeleton[skel_index + 1][0], skeleton[skel_index + 1][1]),
                 color=color, thickness=1)
    cv2.circle(image, (skeleton[0][0], skeleton[0][1]), 2, head_color, -1)


def draw_measurements(image, skeleton, measurements, color):

    if np.any(np.isnan(skeleton)):
        return

    head_joint, mid_body_joint, tail_body_joint = get_joint_indexes(len(skeleton))

    cv2.circle(image, center=(int(skeleton[head_joint][0]), int(skeleton[head_joint][1])),
               radius=int(measurements['head_width'] / 2), color=color)
    cv2.circle(image, center=(int(skeleton[mid_body_joint][0]), int(skeleton[mid_body_joint][1])),
               radius=int(measurements['midbody_width'] / 2), color=color)
    cv2.circle(image, center=(int(skeleton[tail_body_joint][0]), int(skeleton[tail_body_joint][1])),
               radius=int(measurements['tail_width'] / 2), color=color)