import logging
import os
from contextlib import ExitStack

import numpy as np

from wormpose.machine_learning import tfrecord_file
from wormpose.images.real_dataset import RealDataset
from wormpose.pose.worm_skeleton_utils import skeleton_to_angle

logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def generate(frame_preprocessing, dataset_out_dir, frames_dataset, features_dataset, image_shape, num_samples, theta_dims):
    labelled_frames = {}
    for video_name in frames_dataset.video_names():
        skel_is_not_nan = ~np.any(np.isnan(features_dataset[video_name].skeletons), axis=(1, 2))
        template_indexes = np.where(skel_is_not_nan)[0]
        if len(template_indexes) > 0:
            labelled_frames[video_name] = template_indexes

    if len(labelled_frames) == 0:
        raise RuntimeError("Can't create evaluation data because couldn't find any labelled frame in the dataset.")

    len_labelled_frames = int(np.sum([len(x) for x in labelled_frames.values()]))
    if len_labelled_frames < num_samples:
        logging.warning("Not enough labelled frames in the dataset "
                        "to create an evaluation set of {} samples, "
                        "using all available {} samples instead.".format(num_samples, len_labelled_frames))
        num_samples = len_labelled_frames

    real_dataset = RealDataset(frame_preprocessing, image_shape)

    tfrecord_filename = os.path.join(dataset_out_dir, "eval.tfrecord")

    with tfrecord_file.Writer(tfrecord_filename) as record_writer:

        with ExitStack() as stack:
            frames = {v: stack.enter_context(frames_dataset.open(v)) for v in frames_dataset.video_names()}

            for _ in range(num_samples):
                chosen_video_name, chosen_frame_index = _choose_random_frame(labelled_frames)

                image_data, _ = real_dataset.process_frame(frames[chosen_video_name][chosen_frame_index])

                cur_skel = features_dataset[chosen_video_name].skeletons[chosen_frame_index]
                cur_theta = skeleton_to_angle(cur_skel, theta_dims=theta_dims)
                cur_theta_flipped = np.flip(cur_theta, axis=0) + np.pi

                record_writer.write(image_data, cur_theta, cur_theta_flipped)

    return num_samples


def _choose_random_frame(labelled_frames):
    """
    pick randomly a video name and a valid frame index for that video, without repetition
    """
    chosen_video_name = np.random.choice(list(labelled_frames.keys()))
    valid_index_in_video = labelled_frames[chosen_video_name]
    pick_index_at = np.random.randint(0, len(valid_index_in_video))
    chosen_frame_index = valid_index_in_video[pick_index_at]
    # remove chosen frame from the available frames
    labelled_frames[chosen_video_name] = np.delete(labelled_frames[chosen_video_name], pick_index_at)
    if len(labelled_frames[chosen_video_name]) == 0:
        del labelled_frames[chosen_video_name]
    return chosen_video_name, chosen_frame_index
