import os
import pickle
from multiprocessing import Process, Manager

import math
import numpy as np

from wormpose.images.real_dataset import RealDataset


def _read_data(data_reading_queue, num_frames, num_consumer_process, batch_size, temp_dir, frames_dataset, video_name):
    with frames_dataset.open(video_name) as frames:
        num_chunks = math.ceil(float(num_frames) / batch_size)
        for chunk_index in range(num_chunks):
            start = chunk_index * batch_size
            end = min(num_frames, start + batch_size)
            chunk_data = frames[start:end]

            data_filename = os.path.join(temp_dir, "data_to_predict_{}.npy".format(chunk_index))
            with open(data_filename, 'wb') as f:
                pickle.dump(chunk_data, f)

            data_reading_queue.put((data_filename, chunk_index))

    for _ in range(num_consumer_process):
        data_reading_queue.put(None)


def _assemble_images_batch(frame_preprocessing, data_reading_queue, results_queue, temp_dir, batch_size, image_shape):
    images_data_batch = np.zeros(([batch_size, image_shape[0], image_shape[1], 1]), dtype=np.float32)

    real_dataset = RealDataset(frame_preprocessing, image_shape)
    while True:
        queue_data = data_reading_queue.get()
        if queue_data is None:
            break

        data_filename, chunk_index = queue_data
        with open(data_filename, 'rb') as f:
            chunk_data = pickle.load(f)
        os.remove(data_filename)

        for index_in_batch in range(len(chunk_data)):
            cur_frame, _ = real_dataset.process_frame(chunk_data[index_in_batch])
            images_data_batch[index_in_batch, :, :, 0] = cur_frame

        images_data_batch /= 255.

        image_filename = os.path.join(temp_dir, "real_topredict_{:09d}.npy".format(chunk_index))
        np.save(image_filename, images_data_batch)
        results_queue.put(image_filename)


def run(frame_preprocessing, num_frames, frames_dataset, video_name, num_process, temp_dir, image_shape, batch_size):
    manager = Manager()
    data_reading_queue = manager.Queue()
    results_queue = manager.Queue()

    workers = [Process(target=_assemble_images_batch,
                       args=(frame_preprocessing, data_reading_queue, results_queue, temp_dir, batch_size, image_shape))
               for _ in range(num_process)]
    for w in workers:
        w.start()

    _read_data(data_reading_queue=data_reading_queue,
               num_frames=num_frames,
               num_consumer_process=num_process,
               batch_size=batch_size,
               temp_dir=temp_dir,
               frames_dataset=frames_dataset,
               video_name=video_name)

    for w in workers:
        w.join()

    results_files = []
    while not results_queue.empty():
        results_files.append(results_queue.get())
    results_files = sorted(results_files)

    for f in results_files:
        batch_data = np.load(f)
        yield batch_data
        os.remove(f)
