import os
import pickle
from contextlib import ExitStack
from functools import partial
from multiprocessing import Process, Manager

import numpy as np


def _read_data(temp_dir, num_process, video_names, all_template_indexes,
               features_dataset, frames_dataset, max_template_frames=2000):
    """
    Split a subset of template frames and associated features from different videos randomly to several processes
    """
    from contextlib import ExitStack

    projected_num_templates_per_video = max(max_template_frames // len(video_names) // num_process, 1)
    actual_num_templates_per_video = [min(projected_num_templates_per_video, len(x)) for x in all_template_indexes]
    num_total_templates = np.sum(actual_num_templates_per_video)

    template_frames = None
    template_measurements = None
    template_skeletons = None
    template_video_names = None

    with ExitStack() as stack:
        files_frames = {v: stack.enter_context(frames_dataset.open(v)) for v in video_names}

        for process_index in range(num_process):
            cur_index = 0
            for video_name, cur_video_template_indexes, cur_video_num_templates \
                    in zip(video_names, all_template_indexes, actual_num_templates_per_video):

                if cur_video_num_templates == 0:
                    continue

                features = features_dataset[video_name]
                frames = files_frames[video_name]

                template_indexes = np.sort(np.random.choice(cur_video_template_indexes,
                                                            size=cur_video_num_templates,
                                                            replace=False))
                if template_frames is None:
                    template_frames = np.empty((num_total_templates,) + frames.shape[1:],
                                               dtype=frames.dtype)
                if template_skeletons is None:
                    template_skeletons = np.empty((num_total_templates,) + features.skeletons.shape[1:],
                                                  dtype=features.skeletons.dtype)
                if template_video_names is None:
                    template_video_names = np.empty((num_total_templates,), dtype=object)
                if template_measurements is None:
                    template_measurements = np.empty((num_total_templates,) + features.measurements.shape[1:],
                                                     dtype=features.measurements.dtype)

                for i in range(cur_video_num_templates):
                    template_index = template_indexes[i]
                    template_frames[cur_index + i] = frames[template_index]
                    template_skeletons[cur_index + i] = features.skeletons[template_index]
                    template_measurements[cur_index + i] = features.measurements[template_index]
                    template_video_names[cur_index + i] = video_name

                cur_index += cur_video_num_templates

            template_filename = os.path.join(temp_dir, 'template_frames_{:02d}.pkl'.format(process_index))
            with open(template_filename, 'wb') as f:
                pickle.dump({'frames': template_frames,
                             'skeletons': template_skeletons,
                             'measurements': template_measurements,
                             'video_names': template_video_names}
                            , f)
            yield template_filename


def _write_to_file(index,
                   file_extension,
                   filename_prefix,
                   frame_preprocessing,
                   dataset_out_dir,
                   num_samples,
                   template_filename,
                   image_shape,
                   postures_model,
                   progress_queue,
                   disable_image_augmentation,
                   save_templates):
    from wormpose.images.synthetic_dataset import SyntheticDataset
    from wormpose.machine_learning import tfrecord_file

    np.random.seed(index)

    with open(template_filename, 'rb') as f:
        templates = pickle.load(f)
        template_frames = templates['frames']
        template_skeletons = templates['skeletons']
        template_measurements = templates['measurements']
        template_video_names = templates['video_names']
    os.remove(template_filename)

    worm_measurements = {}
    for video_name in np.unique(template_video_names):
        indexes = np.where(template_video_names == video_name)[0]
        worm_measurements[video_name] = template_measurements[indexes]

    synthetic_dataset = SyntheticDataset(frame_preprocessing=frame_preprocessing,
                                         output_image_shape=image_shape,
                                         enable_random_augmentations=not disable_image_augmentation)

    gen_postures = postures_model.generate()

    filename = os.path.join(dataset_out_dir, "{}{}.{}".format(filename_prefix, index, file_extension))

    headtail_choice = np.random.choice([0, 1], num_samples)  # switch head tail for data augmentation

    image_data = np.empty(image_shape, dtype=np.uint8)
    total_amount_templates = len(template_frames)

    REPORT_EVERY = 500

    open_file = {'tfrecord': partial(tfrecord_file.Writer, filename),
                 'pkl': partial(open, filename, 'wb')}[file_extension]
    write_file = {'tfrecord': lambda f, data: f.write(*data),
                  'pkl': lambda f, data: pickle.dump(data, f)}[file_extension]

    templates_filename = os.path.join(dataset_out_dir, "{}_templates.{}".format(index, 'pkl'))

    with ExitStack() as stack:
        f = stack.enter_context(open_file())
        if save_templates:
            indexes_f = stack.enter_context(open(templates_filename, "wb"))

        for index in range(num_samples):
            theta = next(gen_postures)
            theta_flipped = np.flip(theta, axis=0) + np.pi
            label_data = theta, theta_flipped
            cur_template_index = np.random.randint(total_amount_templates)
            video_name = template_video_names[cur_template_index]
            template_frame = template_frames[cur_template_index]
            template_skeleton = template_skeletons[cur_template_index]
            synthetic_dataset.generate(theta=label_data[headtail_choice[index]],
                                       template_frame=template_frame,
                                       template_skeleton=template_skeleton,
                                       worm_measurements=worm_measurements[video_name],
                                       out_image=image_data)

            if index > 0 and index % REPORT_EVERY == 0:
                progress_queue.put(REPORT_EVERY)

            write_file(f, (image_data, label_data[0], label_data[1]))

            if save_templates:
                pickle.dump((video_name, template_frame, template_skeleton), indexes_f)

        progress_queue.put(None)


def generate(frame_preprocessing,
             frames_dataset,
             features_dataset,
             num_process,
             dataset_out_dir,
             temp_dir,
             image_shape,
             num_samples,
             postures_model,
             disable_image_augmentation,
             file_extension,
             filename_prefix="",
             save_templates=False,
             video_names=None):

    if video_names is None:
        video_names = frames_dataset.video_names()

    workers = []
    all_template_indexes = []

    for video_name in video_names:
        features = features_dataset[video_name]
        skel_is_not_nan = ~np.any(np.isnan(features.skeletons), axis=(1, 2))
        template_indexes = np.where(skel_is_not_nan)[0]
        all_template_indexes.append(template_indexes)

    if np.sum([len(x) for x in all_template_indexes]) == 0:
        raise RuntimeError("Can't create training data because couldn't find any labelled frame in the dataset.")

    manager = Manager()
    progress_queue = manager.Queue()

    num_samples_per_process = [num_samples // num_process] * num_process
    num_samples_per_process[0] += num_samples % num_process

    for index, template_filename in \
            enumerate(_read_data(temp_dir,
                                 num_process,
                                 video_names,
                                 all_template_indexes,
                                 features_dataset,
                                 frames_dataset)):
        proc = Process(target=_write_to_file, args=(index,
                                                    file_extension,
                                                    filename_prefix,
                                                    frame_preprocessing,
                                                    dataset_out_dir,
                                                    num_samples_per_process[index],
                                                    template_filename,
                                                    image_shape,
                                                    postures_model,
                                                    progress_queue,
                                                    disable_image_augmentation,
                                                    save_templates))
        proc.start()
        workers.append(proc)

    num_entries_written = 0
    num_process_done = 0
    while True:
        queue_data = progress_queue.get()
        if queue_data is None:
            num_process_done += 1
            if num_process_done == num_process:
                break
        else:
            num_entries_written += queue_data
            yield float(num_entries_written) / num_samples

    for proc in workers:
        proc.join()
