from functools import partial

import tensorflow as tf


def parse_example(example, theta_dims):
    features = {
        'data': tf.io.FixedLenFeature([], tf.string),
        'label0': tf.io.FixedLenFeature([theta_dims], tf.float32),
        'label1': tf.io.FixedLenFeature([theta_dims], tf.float32),
    }
    parsed_features = tf.io.parse_single_example(example, features)
    return parsed_features


def parse_example_normalize_image(example, theta_dims, image_shape):
    parsed_features = parse_example(example, theta_dims)

    img = tf.io.decode_raw(parsed_features['data'], tf.uint8)
    img = tf.reshape(img, (image_shape[0], image_shape[1]))
    img = img[:, :, tf.newaxis]
    img = tf.cast(img, tf.float32)
    img /= 255.

    labels = [parsed_features['label0'], parsed_features['label1']]

    return img, labels


class Writer(object):

    def __init__(self, filename):
        self.filename = filename
        self.record_writer = None

    def __enter__(self):
        options = tf.io.TFRecordOptions(compression_type='GZIP')
        self.record_writer = tf.io.TFRecordWriter(self.filename, options=options)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.record_writer.close()

    def write(self, image_data, label_0, label_1):
        feature = {
            'data': tf.train.Feature(bytes_list=tf.train.BytesList(value=[image_data.tostring()])),
            'label0': tf.train.Feature(float_list=tf.train.FloatList(value=label_0.tolist())),
            'label1': tf.train.Feature(float_list=tf.train.FloatList(value=label_1.tolist()))
        }
        example_proto = tf.train.Example(features=tf.train.Features(feature=feature))
        self.record_writer.write(example_proto.SerializeToString())


def parse_square_image(example, theta_dims):

    parsed_features = parse_example(example, theta_dims)
    img = tf.io.decode_raw(parsed_features['data'], tf.uint8)
    img_size = tf.math.sqrt(tf.cast(tf.shape(img)[0], tf.float32))
    img = tf.reshape(img, (img_size, img_size))
    return img, [parsed_features['label0'], parsed_features['label1']]


def read(filename, theta_dims):
    """
    This function only works when eager execution is enabled, the images in the files should have the same width and height
    """
    raw_dataset = tf.data.TFRecordDataset(filename, compression_type='GZIP')
    parsed_dataset = raw_dataset.map(partial(parse_square_image, theta_dims=theta_dims))

    for parsed_record in parsed_dataset:
        yield parsed_record
