import logging
import math

import numpy as np
import numpy.ma as ma

logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

CONTINUOUS_ANGLES_DIST_THRESHOLD = np.deg2rad(5)
CONTINOUS_SEGMENT_TIME_WINDOW_SEC = 0.2
MIN_SEGMENT_SIZE_SEC = 0.2


def _assemble_continuous_partitions(shuffled_theta,
                                    shuffled_scores,
                                    shuffled_skeletons,
                                    score_threshold,
                                    frame_rate,
                                    dist_threshold=CONTINUOUS_ANGLES_DIST_THRESHOLD):
    """
    Gather continuous segments with consistent head-tail orientation
    """
    time_window = int(frame_rate * CONTINOUS_SEGMENT_TIME_WINDOW_SEC)
    min_segment_size = int(frame_rate * MIN_SEGMENT_SIZE_SEC)

    partitioned_theta = ma.masked_all_like(shuffled_theta)
    partitioned_skeletons = ma.masked_all_like(shuffled_skeletons)
    partitioned_scores = ma.masked_all_like(shuffled_scores)
    partitions = ma.masked_all((len(shuffled_theta),), dtype=np.int64)
    cur_partition = -1

    def decide_partition(frame_index, partition, partition_index):
        partitioned_theta[frame_index][0] = shuffled_theta[frame_index, partition]
        partitioned_theta[frame_index][1] = shuffled_theta[frame_index, 1 - partition]
        partitioned_skeletons[frame_index][0] = shuffled_skeletons[frame_index, partition]
        partitioned_skeletons[frame_index][1] = shuffled_skeletons[frame_index, 1 - partition]
        partitioned_scores[frame_index][0] = shuffled_scores[frame_index][partition]
        partitioned_scores[frame_index][1] = shuffled_scores[frame_index][1 - partition]
        partitions[frame_index] = partition_index

    # discard low score frames
    good_score_frames = np.where(ma.max(shuffled_scores, axis=1) >= score_threshold)[0]

    for frame_index in good_score_frames:

        prev_theta = partitioned_theta[frame_index - min(time_window, frame_index):frame_index, 0]

        # if there is a big gap > time_window we start a new partition
        if np.all(np.any(prev_theta.mask, axis=1)):
            cur_partition += 1
            decide_partition(frame_index=frame_index, partition=0, partition_index=cur_partition)
        # otherwise we look in the time_window close past the closest non nan frame see if we can continue the
        # partition as long as the values stay continuous
        else:
            last_valid_index = np.where(~np.any(prev_theta.mask, axis=1))[0][-1]
            dists = [_angle_distance(shuffled_theta[frame_index, k, :], prev_theta[last_valid_index]) for k in range(2)]
            partition = int(np.argmin(dists))
            if dists[partition] < dist_threshold:
                decide_partition(frame_index=frame_index, partition=partition, partition_index=cur_partition)

    # remove short segments < min_segment_size
    segments = []
    for partition_index in range(cur_partition + 1):
        cur_partition_indices = np.where(partitions == partition_index)[0]
        if len(cur_partition_indices) < min_segment_size:
            partitioned_theta.mask[cur_partition_indices] = True
            partitioned_skeletons[cur_partition_indices] = True
            partitioned_scores[cur_partition_indices] = True
        else:
            segments.append(cur_partition_indices)

    return partitioned_theta, partitioned_skeletons, partitioned_scores, segments


def _angle_distance(angle_a, angle_b):
    return np.linalg.norm(np.arctan2(np.sin(angle_b - angle_a), np.cos(angle_b - angle_a))) / float(len(angle_a))


def skeleton_distance(skel_a, skel_b):
    def head_tail_diff(skel):
        return skel[-1][0] - skel[0][0], skel[-1][1] - skel[0][1]

    def cos_similarity(a, b):
        norm_a = math.sqrt(a[0] * a[0] + a[1] * a[1])
        norm_b = math.sqrt(b[0] * b[0] + b[1] * b[1])
        return (a[0] * b[0] + a[1] * b[1]) / (norm_a * norm_b)

    return cos_similarity(head_tail_diff(skel_a), head_tail_diff(skel_b))


def _align_segments_with_labels(segments, partitioned_skeletons, labelled_skeletons, min_labelled=5):
    """
    Match the head/tail alignment with the results of the classical tracking in each of the segments,
     if there is enough labelled data in the segment
    """
    segments_alignment = ma.masked_all((len(segments),), dtype=np.uint8)
    for segment_index, segment in enumerate(segments):
        segment_skeletons = labelled_skeletons[segment]
        non_nan_labelled = np.any(~np.isnan(segment_skeletons), axis=(1, 2))
        labels_count = np.sum(non_nan_labelled)
        non_masked = ~np.any(partitioned_skeletons[segment].mask, axis=(1, 2, 3))
        to_compare = np.logical_and(non_nan_labelled, non_masked)

        similarity_scores = []
        for label_skel, partitioned_skeleton \
                in zip(segment_skeletons[to_compare], partitioned_skeletons[segment][to_compare]):
            dists = [skeleton_distance(label_skel, x) for x in partitioned_skeleton]
            similarity_scores.append(dists)

        if len(similarity_scores) > 0:
            mean_similarity_scores = np.mean(similarity_scores, axis=0)
            if mean_similarity_scores[0] * mean_similarity_scores[1] < 0 and labels_count > min_labelled:
                segments_alignment[segment_index] = np.argmax(mean_similarity_scores)

    return segments_alignment


def _calculate_smallest_gap_to_adjacent(segment_index, segments, segments_alignment):
    # evaluate how far away this segment is from known values
    score = np.nan
    segment_offset = np.nan
    if segment_index - 1 >= 0 and not segments_alignment.mask[segment_index - 1]:
        gap = segments[segment_index][0] - segments[segment_index - 1][-1]
        score = gap
        segment_offset = -1
    if segment_index + 1 < len(segments_alignment) and not segments_alignment.mask[segment_index + 1]:
        gap = segments[segment_index + 1][0] - segments[segment_index][-1]
        if np.isnan(score) or gap < score:
            score = gap
            segment_offset = 1

    return score, segment_offset


def _align_segment_with_neighbor(segment_to_fix, segment_to_compare_offset, segments, segments_alignment,
                                 partitioned_skeletons):
    cur_segment = segments[segment_to_fix]
    if segment_to_compare_offset == -1:
        # compare first frame of cur segment to last valid frame of previous segment
        prev_segment = segments[segment_to_fix - 1]
        prev_segment_skeleton = partitioned_skeletons[prev_segment[0]:prev_segment[1]][:,
                                segments_alignment[segment_to_fix - 1]]
        prev_valid_frame = np.where(~np.any(prev_segment_skeleton.mask, axis=(1, 2)))[0][-1]
        prev_frame_skeleton = prev_segment_skeleton[prev_valid_frame]
        cur_frame_unaligned_skeleton = partitioned_skeletons[cur_segment[0]]
        dists = [skeleton_distance(prev_frame_skeleton, x) for x in cur_frame_unaligned_skeleton]

    elif segment_to_compare_offset == 1:
        # compare last frame of cur segment to first valid frame of next segment
        next_segment = segments[segment_to_fix + 1]
        next_segment_skeleton = partitioned_skeletons[next_segment[0]:next_segment[1]][:,
                                segments_alignment[segment_to_fix + 1]]
        next_valid_frame = np.where(~np.any(next_segment_skeleton.mask, axis=(1, 2)))[0][0]
        next_frame_skeleton = next_segment_skeleton[next_valid_frame]
        cur_frame_unaligned_skeleton = partitioned_skeletons[cur_segment[-1]]
        dists = [skeleton_distance(next_frame_skeleton, x) for x in cur_frame_unaligned_skeleton]
    else:
        raise ValueError()

    cur_segment_alignement = int(np.argmax(dists))
    segments_alignment[segment_to_fix] = cur_segment_alignement


def _align_unlabelled_segments_with_adjacents(segments, segments_alignment, partitioned_skeletons,
                                              frame_rate):
    """
    Resolve the unaligned segments by comparing with adjacent segments,
    starting with the segments that have the least frames gap between an adjacent trusted segment
    Don't align isolated segments which a big gap between trusted segments
    """
    MAXIMUM_GAP_SECONDS = 1
    maximum_gap_allowed = int(frame_rate * MAXIMUM_GAP_SECONDS)
    # ensure that if no segments have been aligned at all, pick one solution randomly to start
    if np.all(segments_alignment.mask):
        logger.info("There are no trusted segments with head decision to resolve the whole video, stopping analysis.")
        return segments_alignment

    # fix in priority the segments with known adjacent frames with little gap
    # until all segments are aligned except the isolated ones (further than maximum_gap_allowed)
    unaligned = np.where(segments_alignment.mask)[0]
    while len(unaligned) > 0:
        # we first pick the best candidate segment to align (there are known frames nearby before or after or both)
        all_gaps = [_calculate_smallest_gap_to_adjacent(segment_index=x,
                                                        segments=segments,
                                                        segments_alignment=segments_alignment) for x in unaligned]
        segment_to_fix_index = np.nanargmin(all_gaps, axis=0)[0]
        gap_to_adjacent_segment, adjacent_segment_offset = all_gaps[segment_to_fix_index]

        # abort if only isolated segments are left
        if gap_to_adjacent_segment > maximum_gap_allowed:
            break

        _align_segment_with_neighbor(segment_to_fix=unaligned[segment_to_fix_index],
                                     segment_to_compare_offset=adjacent_segment_offset,
                                     segments=segments,
                                     segments_alignment=segments_alignment,
                                     partitioned_skeletons=partitioned_skeletons)

        unaligned = np.where(segments_alignment.mask)[0]

    return segments_alignment


def  _discard_low_scores(score_threshold, partitioned_scores, partitioned_skeletons, partitioned_theta):
    low_scores_index = np.where(partitioned_scores < score_threshold)[0]
    partitioned_scores.mask[low_scores_index] = True
    partitioned_skeletons.mask[low_scores_index] = True
    partitioned_theta.mask[low_scores_index] = True


def resolve_head_tail(shuffled_theta,
                      shuffled_scores,
                      shuffled_skeletons,
                      original_skeletons,
                      frame_rate,
                      score_threshold):
    import numpy.ma as ma
    len_series = len(shuffled_theta)
    # Create continuous segments without jumps
    partitioned_theta, partitioned_skeletons, partitioned_scores, segments \
        = _assemble_continuous_partitions(shuffled_theta=shuffled_theta,
                                          shuffled_scores=shuffled_scores,
                                          shuffled_skeletons=shuffled_skeletons,
                                          score_threshold=score_threshold,
                                          frame_rate=frame_rate)

    if len(segments) == 0:
        logger.error("Couldn't find any continuous segments of predicted data above the threshold {},"
                     " stopping analysis.".format(score_threshold))
        return partitioned_theta[:, 0].filled(np.nan), \
               partitioned_scores[:, 0].filled(np.nan), \
               partitioned_skeletons[:, 0].filled(np.nan)

    # Choose each segment global alignment by comparing with labelled data
    segments_alignment = _align_segments_with_labels(segments, partitioned_skeletons, original_skeletons)

    # Fix unaligned segments here by comparing skeletons with neighboring segments iteratively
    segments_alignment = _align_unlabelled_segments_with_adjacents(segments, segments_alignment, partitioned_skeletons,
                                                                   frame_rate)

    # Compile results
    processed_theta = ma.masked_all((len_series,) + partitioned_theta.shape[2:], dtype=partitioned_theta.dtype)
    processed_skeletons = ma.masked_all((len_series,) + partitioned_skeletons.shape[2:],
                                        dtype=partitioned_skeletons.dtype)
    processed_scores = ma.masked_all((len_series,), dtype=partitioned_scores.dtype)
    for segment, segment_alignment in zip(segments, segments_alignment):
        if not ma.is_masked(segment_alignment):
            processed_scores[segment] = partitioned_scores[segment][:, segment_alignment]
            processed_skeletons[segment] = partitioned_skeletons[segment][:, segment_alignment]
            processed_theta[segment] = partitioned_theta[segment][:, segment_alignment]

    # Filter the final results again by score threshold
    _discard_low_scores(score_threshold, partitioned_scores, partitioned_skeletons, partitioned_theta)

    num_success = np.sum(~processed_scores.mask)
    total = len(processed_scores)
    original_num_success = np.any(~np.isnan(original_skeletons), axis=(1, 2)).sum()
    logger.info("Resolved head/tail, {} out of {} frames "
                "analyzed successfully ({:.1f}%) "
                "(original features : {} or {:.1f}% of total)".format(num_success,
                                                                      total,
                                                                      float(num_success) / total * 100,
                                                                      original_num_success,
                                                                      float(original_num_success) / total * 100))
    if num_success < original_num_success:
        logger.warning(
            "Original results had {} more successfully analyzed frames!".format(original_num_success - num_success))

    return processed_theta.filled(np.nan), processed_scores.filled(np.nan), processed_skeletons.filled(np.nan)
