import json
import logging
import os

import h5py
import numpy as np
import numpy.ma as ma
from scipy.interpolate import splrep, splev
from scipy.signal import savgol_filter

from wormpose import default_config
from wormpose.dataset.loader import load_dataset
from wormpose.images.worm_drawing import calculate_backbone_points

logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def _spline_interpolation(tseries, scores):
    w = 1 / np.abs((1 - scores))
    interpolated_tseries = np.copy(tseries)
    for dim in range(tseries.shape[1]):
        y = tseries[:, dim]
        new_x = np.arange(len(y))
        sel = ~np.isnan(y)
        x = new_x[sel]
        y = y[sel]
        weights = w[sel]
        spl = splrep(x, y, w=weights, s=len(x) + np.sqrt(2 * len(x)))
        new_weighted_y = splev(new_x, spl)
        interpolated_tseries[:, dim] = new_weighted_y
    return interpolated_tseries


def _interpolate_tseries(tseries, scores, missing_data_threshold, frames_around):
    gaps = np.any(np.isnan(tseries), axis=1)
    gaps_boundaries = np.where(np.abs(np.diff(np.concatenate([[False], gaps, [False]]))))[0].reshape(-1, 2)
    t0 = 0
    fixable_segments = []
    for gaps in gaps_boundaries[np.hstack(np.diff(gaps_boundaries, axis=1)) > missing_data_threshold]:
        if gaps[0] - t0 > frames_around:
            fixable_segments.append([t0, gaps[0]])
        t0 = gaps[1]
    if t0 < len(tseries) - frames_around:
        fixable_segments.append([t0, len(tseries)])
    new_tseries = ma.zeros(tseries.shape)
    for t0, tf in fixable_segments:
        new_tseries[t0:tf] = _spline_interpolation(tseries[t0:tf], scores[t0:tf])
    new_tseries[new_tseries == 0] = ma.masked
    return new_tseries.filled(np.nan)


def _smooth_tseries(interp_tseries, window_size, poly_order):
    gaps = np.any(np.isnan(interp_tseries), axis=1)
    segments = np.where(np.abs(np.diff(np.concatenate([[True], gaps, [True]]))))[0].reshape(-1, 2)
    smooth_tseries = np.zeros(interp_tseries.shape)
    for t0, tf in segments:
        smooth_tseries[t0:tf] = savgol_filter(interp_tseries[t0:tf], axis=0,
                                              window_length=window_size, polyorder=poly_order)
    smooth_tseries[np.any(np.isnan(interp_tseries), axis=1)] = np.nan
    return smooth_tseries


def _dorsal_ventral_flip_modes(modes):
    return -modes


def _dorsal_ventral_flip_theta(theta):
    return 2 * np.pi - theta


def _calculate_processed_skeleton(features, flipped, skeletons_raw, theta_smooth):
    average_worm_length = np.nanmean(features.measurements['worm_length'])
    skeletons_smooth = []
    raw_skel_not_nan = np.where(~np.any(np.isnan(skeletons_raw), axis=(1, 2)))[0]
    for index, t in enumerate(theta_smooth if not flipped else _dorsal_ventral_flip_theta(theta_smooth)):
        rx, ry = calculate_backbone_points(t, average_worm_length)
        rx -= np.mean(rx)
        ry -= np.mean(ry)
        closest_raw_skel_index = raw_skel_not_nan[np.argmin(np.abs(raw_skel_not_nan - index))]
        x_mean, y_mean = np.mean(skeletons_raw[closest_raw_skel_index], axis=0)
        rx += x_mean
        ry += y_mean
        skeletons_smooth.append((rx, ry))
    skeletons_smooth = np.swapaxes(np.array(skeletons_smooth), 1, 2)
    return skeletons_smooth


def post_process(dataset_path,
                 work_dir=None,
                 smoothing_window_seconds=None,
                 data_thresh_seconds=None,
                 max_gap_seconds=None,
                 poly_order=None):
    if work_dir is None:
        work_dir = default_config.WORK_DIR
    if data_thresh_seconds is None:
        data_thresh_seconds = 0.5
    if max_gap_seconds is None:
        max_gap_seconds = 0.2
    if smoothing_window_seconds is None:
        smoothing_window_seconds = 0.5
    if poly_order is None:
        poly_order = 3

    logger.info("Postprocessing with parameters: \n\t" + ', \n\t'.join(['{}: {}'.format(x[0], x[1])
                                                                        for x in {'dataset_path': os.path.abspath(dataset_path),
                                                                                  'work_dir': os.path.abspath(work_dir),
                                                                                  'smoothing_window_seconds': smoothing_window_seconds,
                                                                                  'data_thresh_seconds': data_thresh_seconds,
                                                                                  'max_gap_seconds': max_gap_seconds,
                                                                                  'poly_order': poly_order
                                                                                  }.items()]))

    dataset_root_name = os.path.basename(os.path.normpath(dataset_path))
    project_dir = os.path.join(work_dir, dataset_root_name)
    results_root_dir = os.path.join(project_dir, default_config.RESULTS_DIR)

    eigenworms_matrix_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'resources', 'EigenWorms.csv')
    eigenworms_matrix = np.loadtxt(eigenworms_matrix_path, delimiter=',').astype(np.float32)

    logger.info('Load predictions.')

    with open(os.path.join(project_dir, default_config.CONFIG_FILENAME), 'r') as processed_f:
        config = json.load(processed_f)
    features_dataset, _, _ = load_dataset(config['dataset_loader'], dataset_path)

    with h5py.File(os.path.join(results_root_dir, 'processed_results.h5'), 'w') as processed_f:

        metadata = processed_f.attrs
        metadata['data_thresh_seconds'] = data_thresh_seconds
        metadata['max_gap_seconds'] = max_gap_seconds
        metadata['smoothing_window_seconds'] = smoothing_window_seconds
        metadata['poly_order'] = poly_order

        for video_name in sorted(os.listdir(results_root_dir)):

            if not os.path.isdir(os.path.join(results_root_dir, video_name)):
                continue

            results_file = os.path.join(results_root_dir, video_name, 'results.h5')
            if not os.path.exists(results_file):
                logger.info("No results file to analyze, file not found: {}".format(results_file))
                continue

            frame_rate = features_dataset[video_name].frame_rate
            max_gap_frames = int(max_gap_seconds * frame_rate)
            frames_around = int(data_thresh_seconds * frame_rate)
            smoothing_window_frames = int(smoothing_window_seconds * frame_rate)

            with h5py.File(results_file, 'r') as results_f:

                scores = results_f['scores'][:]

                theta_raw = results_f['theta'][:]
                mean_theta_raw = np.mean(theta_raw, axis=1)
                modes_raw = (theta_raw - mean_theta_raw[:, np.newaxis]).dot(eigenworms_matrix)
                theta_orig = results_f['original_theta'][:]
                skeletons_raw = results_f['skeletons'][:]

                features = features_dataset[video_name]

                modes_interp = _interpolate_tseries(modes_raw, scores, max_gap_frames, frames_around)
                modes_smooth = _smooth_tseries(modes_interp, smoothing_window_frames, poly_order)

                flipped = False
                if features.ventral_side == 'clockwise':
                    logger.info('Flipping dorsal-ventral side.')
                    modes_raw = _dorsal_ventral_flip_modes(modes_raw)
                    modes_interp = _dorsal_ventral_flip_modes(modes_interp)
                    modes_smooth = _dorsal_ventral_flip_modes(modes_smooth)
                    mean_theta_raw = _dorsal_ventral_flip_theta(mean_theta_raw)
                    flipped = True

                theta_interp = eigenworms_matrix.dot(modes_interp.T).T + mean_theta_raw[:, np.newaxis]
                theta_smooth = eigenworms_matrix.dot(modes_smooth.T).T + mean_theta_raw[:, np.newaxis]

                worm_group = processed_f.create_group(video_name)
                worm_group.create_dataset('modes_raw', data=modes_raw)
                worm_group.create_dataset('modes_interp', data=modes_interp)
                worm_group.create_dataset('modes_smooth', data=modes_smooth)
                worm_group.create_dataset('frame_rate', data=frame_rate)
                worm_group.create_dataset('dorsal_ventral_flip', data=flipped)

                worm_group.create_dataset('theta_raw', data=theta_raw)
                worm_group.create_dataset('theta_interp', data=theta_interp)
                worm_group.create_dataset('theta_smooth', data=theta_smooth)
                worm_group.create_dataset('theta_orig', data=theta_orig)

                skeletons_interp = _calculate_processed_skeleton(features, flipped, skeletons_raw, theta_interp)
                skeletons_smooth = _calculate_processed_skeleton(features, flipped, skeletons_raw, theta_smooth)

                worm_group.create_dataset('skeletons_raw', data=skeletons_raw)
                worm_group.create_dataset('skeletons_interp', data=skeletons_interp)
                worm_group.create_dataset('skeletons_smooth', data=skeletons_smooth)

                logger.info('Post-processed worm {}'.format(video_name))


def main():
    import argparse

    parser = argparse.ArgumentParser()

    parser.add_argument('dataset_path', type=str)
    parser.add_argument('--work_dir', type=str,
                        help="Root folder for all experiments")
    parser.add_argument('--data_thresh_seconds', type=float,
                        help="minimum amount of time (seconds) before gap that needs interpolation")
    parser.add_argument('--max_gap_seconds', type=float,
                        help="maximum amount of time (seconds) allowed to interpolate over")
    parser.add_argument('--smoothing_window_seconds', type=float,
                        help="smoothing window in seconds")
    parser.add_argument('--poly_order', type=int,
                        help="polynomial order in smoothing")
    args = parser.parse_args()

    post_process(**vars(args))


if __name__ == "__main__":
    main()