import json
import logging
import multiprocessing as mp
import os
import shutil
import tempfile
import time
import uuid
from functools import partial

import h5py
import numpy as np
import tensorflow as tf

from wormpose import default_config
from wormpose.images import image_score_generator
from wormpose.machine_learning import predict_data_generator
from wormpose.machine_learning.best_models_saver import BestModels
from wormpose.dataset.loader import load_dataset
from wormpose.pose.headtail_resolution import resolve_head_tail
from wormpose.pose.worm_skeleton_utils import skeletons_to_angles

logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
tf.get_logger().setLevel(logging.INFO)


def _resample(values, timestamp):
    values = np.array(values)
    resampled = np.full((timestamp[-1] - timestamp[0] + 1,) + values[0].shape, np.nan, values.dtype)
    for i in range(len(values)):
        t = timestamp[i] - timestamp[0]
        resampled[t] = values[i]
    return resampled


def network_predict(frame_preprocessing,
                    video_name,
                    frames_dataset,
                    num_process,
                    temp_dir,
                    image_shape,
                    batch_size,
                    model_path,
                    num_frames=None):
    tf.compat.v1.reset_default_graph()

    if not os.path.isfile(model_path):
        raise FileNotFoundError("Couldn't find the trained model ({}), stopping analysis.".format(model_path))

    keras_model = tf.keras.models.load_model(model_path, compile=False)

    with frames_dataset.open(video_name) as frames:
        if num_frames is None:
            num_frames = len(frames)
        else:
            num_frames = min(num_frames, len(frames))

    predict_data_gen = partial(predict_data_generator.run,
                               frame_preprocessing=frame_preprocessing,
                               frames_dataset=frames_dataset,
                               num_frames=num_frames,
                               video_name=video_name,
                               num_process=num_process,
                               temp_dir=temp_dir,
                               image_shape=image_shape,
                               batch_size=batch_size)

    def predict_input_fn():
        im_shape = [batch_size, image_shape[0], image_shape[1], 1]
        dset = tf.data.Dataset.from_generator(predict_data_gen, tf.float32, tf.TensorShape(im_shape))
        return dset

    start = time.time()

    predictions = keras_model.predict(predict_input_fn())

    end = time.time()

    predictions = predictions[:num_frames]

    logger.info("Predicted \"{}\" ({} frames) in {:.1f}s".format(video_name, len(predictions), end - start))
    return predictions


def setup_two_solutions(network_predictions):
    """
    Convert network predictions to contain the two solutions (head-tail flip) for each frame
    """
    return np.concatenate([network_predictions[:, np.newaxis],
                    np.flip(network_predictions, axis=1)[:, np.newaxis] + np.pi], axis=1)


def calculate_scores(frame_preprocessing, video_name, frames_dataset, theta_two_options, features_dataset, temp_dir,
                     num_process, image_shape):
    start = time.time()
    matching_scores, matching_skeletons = image_score_generator.run(frame_preprocessing=frame_preprocessing,
                                                                    temp_dir=temp_dir,
                                                                    num_process=num_process,
                                                                    video_name=video_name,
                                                                    frames_dataset=frames_dataset,
                                                                    features_dataset=features_dataset,
                                                                    theta_two_options=theta_two_options,
                                                                    image_shape=image_shape)
    end = time.time()
    logger.info(
        "Calculated scores in {:.1f}s, "
        "average error metric: {:.4f}".format(end - start, np.max(matching_scores, axis=1).mean()))

    return matching_scores, matching_skeletons


def predict(dataset_path,
            scratch_dir=None,
            model_path=None,
            work_dir=None,
            num_process=None,
            batch_size=None,
            score_threshold=None):
    mp.set_start_method('spawn', force=True)

    if work_dir is None:
        work_dir = default_config.WORK_DIR

    dataset_root_name = os.path.basename(os.path.normpath(dataset_path))
    project_dir = os.path.join(work_dir, dataset_root_name)
    if model_path is None:
        best_models_callback = BestModels(os.path.join(project_dir, default_config.MODELS_DIRS))
        model_path = best_models_callback.best_model_path
    if scratch_dir is None:
        scratch_dir = tempfile.gettempdir()
    if num_process is None:
        num_process = os.cpu_count()
    if batch_size is None:
        batch_size = 512
    if score_threshold is None:
        score_threshold = 0.7

    logger.info("Predicting with parameters: \n\t" + ', \n\t'.join(['{}: {}'.format(x[0], x[1])
                                                                    for x in {'dataset_path': os.path.abspath(dataset_path),
                                                                              'work_dir': os.path.abspath(work_dir),
                                                                              'scratch_dir': os.path.abspath(scratch_dir),
                                                                              'model_path': os.path.abspath(model_path),
                                                                              'num_process': num_process,
                                                                              'batch_size': batch_size,
                                                                              'score_threshold': score_threshold,
                                                                              }.items()]))

    temp_dir = os.path.join(scratch_dir, str(uuid.uuid4()))
    if os.path.exists(temp_dir):
        shutil.rmtree(temp_dir)
    os.mkdir(temp_dir)
    logger.info('Using the temporary folder: {}'.format(temp_dir))

    results_root_dir = os.path.join(project_dir, default_config.RESULTS_DIR)
    if not os.path.exists(results_root_dir):
        os.mkdir(results_root_dir)

    config_file = os.path.join(project_dir, default_config.CONFIG_FILENAME)
    with open(config_file, 'r') as f:
        config = json.load(f)
    image_shape = config['image_shape']
    dataset_loader = config['dataset_loader']

    features_dataset, frames_dataset, frame_preprocessing = load_dataset(dataset_loader, dataset_path)
    video_names = frames_dataset.video_names()

    results_folders = []
    for video_name in video_names:
        results_folder = os.path.join(results_root_dir, video_name)
        if not os.path.exists(results_folder):
            os.mkdir(results_folder)
        results_folders.append(results_folder)

    for video_name, results_folder in zip(video_names, results_folders):
        logger.info("Processing {}".format(video_name))

        features = features_dataset[video_name]

        # run all frames through the neural network to get a result theta without head/tail decision

        network_predictions = network_predict(frame_preprocessing=frame_preprocessing,
                                              video_name=video_name,
                                              frames_dataset=frames_dataset,
                                              num_process=num_process,
                                              temp_dir=temp_dir,
                                              image_shape=image_shape,
                                              batch_size=batch_size,
                                              model_path=model_path)
        theta_dims = network_predictions.shape[1]
        theta_two_options = setup_two_solutions(network_predictions=network_predictions)

        # original data
        resampled_original_skeletons = _resample(features.skeletons, features.timestamp)
        resampled_original_theta = skeletons_to_angles(resampled_original_skeletons, theta_dims=theta_dims)

        # calculate image scores for each frame, for the two solutions
        scores_two_options, skeletons_two_options = calculate_scores(frame_preprocessing=frame_preprocessing,
                                                                     video_name=video_name,
                                                                     frames_dataset=frames_dataset,
                                                                     features_dataset=features_dataset,
                                                                     theta_two_options=theta_two_options,
                                                                     temp_dir=temp_dir,
                                                                     num_process=num_process,
                                                                     image_shape=image_shape)
        if np.all(np.isnan(scores_two_options)):
            logger.error("Calculated scores are all invalid, stopping analysis for {}".format(video_name))
            continue

        if np.max(scores_two_options) < score_threshold:
            logger.error(
                "There is not one frame where the error metric is above the threshold {} "
                "in the whole video {}, stopping analysis. Maybe the model didn't train properly.".format(
                    score_threshold, video_name))
            continue

        # resample all time series then resolve head tail
        scores_two_options = _resample(scores_two_options, features.timestamp)
        theta_two_options = _resample(theta_two_options, features.timestamp)
        skeletons_two_options = _resample(skeletons_two_options, features.timestamp)
        theta_final, scores_final, skeletons_final = \
            resolve_head_tail(theta_two_options, scores_two_options, skeletons_two_options,
                              resampled_original_skeletons,
                              frame_rate=features.frame_rate,
                              score_threshold=score_threshold)

        # save results
        results_filename = 'results.h5'
        results_temp_filepath = os.path.join(temp_dir, results_filename)
        with h5py.File(results_temp_filepath, 'w') as f:
            f.create_dataset('original_theta', data=resampled_original_theta)
            f.create_dataset('original_skeletons', data=resampled_original_skeletons)
            f.create_dataset('theta', data=theta_final)
            f.create_dataset('skeletons', data=skeletons_final)
            f.create_dataset('scores', data=scores_final)

        shutil.move(results_temp_filepath, os.path.join(results_folder, results_filename))

    # cleanup
    shutil.rmtree(temp_dir)


def main():
    import argparse

    parser = argparse.ArgumentParser()

    # model infos
    parser.add_argument('--model_path', type=str, help="Load model from this path, or use best model from work_dir.")
    parser.add_argument('--batch_size', type=int)

    # inputs
    parser.add_argument('dataset_path', type=str)
    parser.add_argument('--scratch_dir', type=str,
                        help="Where to store temporary intermediate results")
    parser.add_argument('--work_dir', type=str,
                        help="Root folder for all experiments")
    # multiprocessing params
    parser.add_argument('--num_process', type=int,
                        help="How many worker processes")
    # parameters of results processing
    parser.add_argument('--score_threshold', type=float,
                        help="Image metric score threshold : discard results scoring lower than this value."
                             " Fine tune this value using the script calibrate.py")
    args = parser.parse_args()

    predict(**vars(args))


if __name__ == "__main__":
    main()