import glob
import json
import logging
import multiprocessing as mp
import os
from functools import partial

import tensorflow as tf

from wormpose import default_config
from wormpose.machine_learning import model, tfrecord_file
from wormpose.machine_learning.best_models_saver import BestModels
from wormpose.machine_learning.loss import symmetric_angle_difference

logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
tf.get_logger().setLevel(logging.INFO)


def get_train_dataset(filenames, image_shape, batch_size, theta_dims):
    dataset = tf.data.TFRecordDataset(filenames, compression_type='GZIP')
    dataset = dataset.shuffle(buffer_size=10000)
    dataset = dataset.repeat()
    dataset = dataset.map(
        partial(tfrecord_file.parse_example_normalize_image, theta_dims=theta_dims, image_shape=image_shape),
        num_parallel_calls=tf.data.experimental.AUTOTUNE)
    dataset = dataset.batch(batch_size)
    dataset = dataset.prefetch(buffer_size=tf.data.experimental.AUTOTUNE)
    return dataset


def get_validation_dataset(filenames, image_shape, batch_size, theta_dims):
    dataset = tf.data.TFRecordDataset(filenames, compression_type='GZIP')
    dataset = dataset.map(
        partial(tfrecord_file.parse_example_normalize_image, theta_dims=theta_dims, image_shape=image_shape),
        num_parallel_calls=tf.data.experimental.AUTOTUNE)
    dataset = dataset.batch(batch_size)
    dataset = dataset.prefetch(buffer_size=tf.data.experimental.AUTOTUNE)
    return dataset


def train(dataset_path,
          batch_size=None,
          work_dir=None,
          network_model=None,
          epochs=None,
          optimizer=None,
          loss=None):
    mp.set_start_method('spawn', force=True)

    if work_dir is None:
        work_dir = default_config.WORK_DIR
    if batch_size is None:
        batch_size = 128
    if epochs is None:
        epochs = 100
    if network_model is None:
        network_model = model.build_model
    if optimizer is None:
        optimizer = 'adam'
    if loss is None:
        loss = symmetric_angle_difference

    logger.info("Training with parameters: \n\t" + ', \n\t'.join(['{}: {}'.format(x[0], x[1])
                                                                  for x in {'dataset_path': os.path.abspath(dataset_path),
                                                                            'batch_size': batch_size,
                                                                            'epochs': epochs,
                                                                            'work_dir': os.path.abspath(work_dir),
                                                                            'network_model': network_model,
                                                                            'loss':loss,
                                                                            'optimizer':optimizer,
                                                                            }.items()]))

    dataset_root_name = os.path.basename(os.path.normpath(dataset_path))
    project_dir = os.path.join(work_dir, dataset_root_name)
    models_dir = os.path.join(project_dir, default_config.MODELS_DIRS)
    if not os.path.exists(models_dir):
        os.mkdir(models_dir)
    logger.info('Saving models in : {}'.format(os.path.abspath(models_dir)))

    # find dataset files for training and evaluation
    training_data_dir = os.path.join(project_dir, default_config.TRAINING_DATA_DIR)
    train_tfrecord_filenames = glob.glob(os.path.join(training_data_dir, "train*.tfrecord"))
    eval_tfrecord_filenames = glob.glob(os.path.join(training_data_dir, "eval.tfrecord"))
    if len(train_tfrecord_filenames) == 0 or len(eval_tfrecord_filenames) == 0:
        raise FileNotFoundError("Training/Eval dataset not found.")

    with open(os.path.join(project_dir, default_config.CONFIG_FILENAME), 'r') as f:
        config = json.load(f)
    image_shape = config['image_shape']
    theta_dims = config['theta_dimensions']
    num_train_samples = config['num_train_samples']
    num_eval_samples = config['num_eval_samples']

    if num_eval_samples < batch_size or num_train_samples < batch_size:
        raise ValueError("The number of samples in the train and eval datasets must be higher than the batch size.")

    tf.compat.v1.reset_default_graph()

    keras_model = network_model(input_shape=image_shape, out_dim=theta_dims)

    train_dataset = get_train_dataset(filenames=train_tfrecord_filenames,
                                                         image_shape=image_shape,
                                                         batch_size=batch_size,
                                                         theta_dims=theta_dims)
    validation_dataset = get_validation_dataset(filenames=eval_tfrecord_filenames,
                                                       image_shape=image_shape,
                                                       batch_size=batch_size,
                                                       theta_dims=theta_dims)

    tensorboard_callback = tf.keras.callbacks.TensorBoard(log_dir= os.path.join(project_dir, 'tensorboard_log'),
                                                          histogram_freq=1)
    best_models_callback = BestModels(models_dir)
    checkpoint_callback = tf.keras.callbacks.ModelCheckpoint(best_models_callback.models_name_pattern,
                                                             save_best_only=False,
                                                             save_weights_only=False,
                                                             monitor='val_loss',
                                                             mode='min')

    last_model_path = best_models_callback.last_model_path
    if os.path.isfile(last_model_path):
        keras_model = tf.keras.models.load_model(last_model_path, compile=False)

    keras_model.compile(optimizer=optimizer, loss=loss)

    steps_per_epoch = num_train_samples // batch_size
    keras_model.fit(train_dataset,
                    epochs=epochs,
                    steps_per_epoch=steps_per_epoch,
                    shuffle=False,
                    initial_epoch=best_models_callback.epoch,
                    validation_data=validation_dataset,
                    validation_steps=num_eval_samples // batch_size,
                    callbacks=[tensorboard_callback, checkpoint_callback, best_models_callback])


def main():
    import argparse

    parser = argparse.ArgumentParser()
    # training params
    parser.add_argument('--batch_size', type=int,
                        help="Batch size for training")
    parser.add_argument('--epochs', type=int,
                        help="How many epochs to train the network")
    parser.add_argument('--work_dir', type=str,
                        help="Root folder for all experiments")
    parser.add_argument('--optimizer', type=str, help="Which optimizer for training, 'adam' by default.")


    # input
    parser.add_argument('dataset_path', type=str)

    args = parser.parse_args()

    train(**vars(args))


if __name__ == "__main__":
    main()