import json
import logging
import os
import shutil
import tempfile
import time
import uuid

import cv2
import h5py
import numpy as np

from wormpose import default_config
from wormpose.dataset.loader import load_dataset
from wormpose.default_config import CONFIG_FILENAME
from wormpose.images.worm_drawing import draw_skeleton

logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def _save_images(features_dataset,
                 frames_dataset,
                 video_name,
                 images_temp_dir,
                 results_dir,
                 results_scores,
                 results_skel,
                 results_skel_processed,
                 draw_original):
    if os.path.exists(images_temp_dir):
        shutil.rmtree(images_temp_dir)
    os.mkdir(images_temp_dir)

    features = features_dataset[video_name]
    timestamp = features.timestamp

    image_filename_format = "time_{{:0{}d}}_frame_{{:0{}d}}_score_{{:.2f}}.png".format(len(str(timestamp[-1])),
                                                                                       len(str(len(results_scores))))

    start = time.time()

    original_color = (255, 255, 255)
    predicted_color = (0, 255, 0)
    processed_color = (0, 0, 255)

    with frames_dataset.open(video_name) as frames:

        for cur_time, (score, skel) in enumerate(zip(results_scores, results_skel)):

            frame_index = np.where(timestamp == cur_time)[0]
            if len(frame_index) == 0:
                continue
            cur_frame_index = frame_index[0]
            cur_raw_frame = cv2.cvtColor(frames[cur_frame_index], cv2.COLOR_GRAY2BGR)

            if draw_original:
                # draw original skeleton in white
                draw_skeleton(cur_raw_frame, features.skeletons[cur_frame_index], original_color, original_color)
            # draw skeleton from network predictions in green
            draw_skeleton(cur_raw_frame, skel, predicted_color, predicted_color)

            if results_skel_processed is not None:
                draw_skeleton(cur_raw_frame, results_skel_processed[cur_frame_index], processed_color, processed_color)

            cv2.imwrite(os.path.join(images_temp_dir, image_filename_format.format(cur_time, cur_frame_index, score)),
                        cur_raw_frame)

    shutil.make_archive(os.path.join(results_dir, 'images_results'), 'zip', images_temp_dir)

    end = time.time()
    logger.info("Exported result images for {} in {:.1f}s".format(os.path.basename(results_dir), end - start))


def visualize(dataset_path,
              scratch_dir=None,
              draw_original=None,
              work_dir=None):
    if work_dir is None:
        work_dir = default_config.WORK_DIR
    if scratch_dir is None:
        scratch_dir = tempfile.gettempdir()
    if draw_original is None:
        draw_original = True

    logger.info("Visualizing with parameters: \n\t" + ', \n\t'.join(['{}: {}'.format(x[0], x[1])
                                                                     for x in {'dataset_path': os.path.abspath(dataset_path),
                                                                               'work_dir': os.path.abspath(work_dir),
                                                                               'scratch_dir': os.path.abspath(scratch_dir),
                                                                               'draw_original': draw_original
                                                                               }.items()]))

    temp_dir = os.path.join(scratch_dir, str(uuid.uuid4()))
    if os.path.exists(temp_dir):
        shutil.rmtree(temp_dir)
    os.mkdir(temp_dir)
    logger.info('Using the temporary folder: {}'.format(temp_dir))
    images_temp_dir = os.path.join(temp_dir, 'images_result')

    dataset_root_name = os.path.basename(os.path.normpath(dataset_path))
    project_dir = os.path.join(work_dir, dataset_root_name)

    results_dir = os.path.join(project_dir, default_config.RESULTS_DIR)

    with open(os.path.join(project_dir, CONFIG_FILENAME), 'r') as f:
        config = json.load(f)
    dataset_loader = config['dataset_loader']

    features_dataset, frames_dataset, _ = load_dataset(dataset_loader, dataset_path)
    video_names = frames_dataset.video_names()

    for video_name in video_names:

        results_file = os.path.join(results_dir, video_name, 'results.h5')
        if not os.path.exists(results_file):
            logger.info("No results file to analyze, file not found: {}".format(results_file))
            continue

        with h5py.File(results_file, 'r') as f:
            scores = f['scores'][:]
            skeletons = f['skeletons'][:]

        processed_results_file = os.path.join(results_dir, 'processed_results.h5')
        if os.path.exists(processed_results_file):
            with h5py.File(processed_results_file, 'r') as f:
                processed_skeletons = f[video_name]['skeletons_smooth'][:]
        else:
            processed_skeletons = None

        _save_images(features_dataset,
                     frames_dataset,
                     video_name,
                     images_temp_dir,
                     os.path.join(results_dir, video_name),
                     scores,
                     skeletons,
                     processed_skeletons,
                     draw_original=draw_original)

    # cleanup
    shutil.rmtree(temp_dir)


def main():
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('dataset_path', type=str)
    parser.add_argument('--scratch_dir', type=str,
                        help="Where to store temporary intermediate results")
    parser.add_argument('--work_dir', type=str,
                        help="Root folder for all experiments")
    args = parser.parse_args()

    visualize(**vars(args))


if __name__ == "__main__":
    main()